from __future__ import annotations

import os
import shutil
import sys
from collections.abc import Iterable
from functools import lru_cache
from itertools import chain
from pathlib import Path


def expand_path(value: os.PathLike[str]) -> Path:
    return Path(value).expanduser().resolve()


def reveal_folder(path: str | os.PathLike[str]) -> None:
    if sys.platform == 'win32':
        os.startfile(path, 'explore')
    else:
        import click

        click.launch(os.fspath(path), locate=True)


@lru_cache(1)
def make_instawowt():
    from tempfile import gettempdir

    instawowt = Path(gettempdir(), 'instawowt')
    instawowt.mkdir(exist_ok=True)
    return instawowt


trash = None

if sys.platform == 'darwin':
    _maybe_trash_bin = shutil.which('/usr/bin/trash')  # Added in macOS 14.
    if _maybe_trash_bin:
        _trash_bin = _maybe_trash_bin

        def _trash_darwin(paths: Iterable[os.PathLike[str]]) -> None:
            import subprocess

            try:
                subprocess.run(
                    [_trash_bin, *(os.fspath(p) for p in paths)], capture_output=True, check=True
                )
            except subprocess.CalledProcessError as error:
                if error.returncode != 5:  # Not file not found
                    error.add_note(error.stderr.decode())
                    raise

        trash = _trash_darwin


if trash is None:

    def _trash_default(paths: Iterable[os.PathLike[str]]) -> None:
        from tempfile import mkdtemp

        paths_iter = iter(paths)
        first_path = next(paths_iter, None)

        if first_path is None:
            return

        instawowt = make_instawowt()
        parent_folder = mkdtemp(dir=instawowt, prefix=f'deleted-{Path(first_path).name}-')

        for path in chain((first_path,), paths_iter):
            try:
                shutil.move(path, parent_folder)
            except FileNotFoundError:
                pass

    trash = _trash_default



from __future__ import annotations

import pytest

from instawow._sources.tukui import TukuiResolver
from instawow.definitions import Defn
from instawow.wow_installations import Track

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


@pytest.fixture
def tukui_resolver():
    return TukuiResolver()


@pytest.mark.parametrize(
    'iw_profile_config_values',
    Track,
    indirect=True,
)
@pytest.mark.parametrize('alias', ['tukui', 'elvui'])
async def test_resolve_addon(
    tukui_resolver: TukuiResolver,
    alias: str,
):
    defn = Defn('tukui', alias)

    result = await tukui_resolver.resolve_one(defn, None)

    assert type(result) is dict
    assert result['slug'] == alias


async def test_changelog_url_format(
    tukui_resolver: TukuiResolver,
):
    defn = Defn('tukui', 'tukui')

    result = await tukui_resolver.resolve_one(defn, None)

    assert result['changelog_url'] == 'https://api.tukui.org/v1/changelog/tukui#20.461'



from __future__ import annotations

from collections.abc import Callable, Sequence


def tabulate(rows: Sequence[tuple[object, ...]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."

    from functools import partial
    from textwrap import fill

    from wcwidth import wcswidth

    truncate = partial(fill, width=max_col_width, max_lines=1)

    def make_col_cells():
        for col in zip(*rows):
            col_cells = [(v, wcswidth(v)) for e in col for v in (truncate(str(e)),)]
            max_width = max(w for _, w in col_cells)
            yield [
                *((v, max_width - w) for v, w in col_cells[:1]),
                ('-' * max_width, 0),
                *((v, max_width - w) for v, w in col_cells[1:]),
            ]

    return '\n'.join('  '.join(f'{v}{" " * w}' for v, w in r) for r in zip(*make_col_cells()))


def normalise_names(replace_delim: str) -> Callable[[str], str]:
    import string

    trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

    def normalise(value: str):
        return replace_delim.join(value.casefold().translate(trans_table).split())

    return normalise



"""
Trimmed down version of aresponses.
"""

from __future__ import annotations

import inspect
import re
from collections.abc import Awaitable, Callable, Sequence
from functools import partial
from typing import Protocol
from unittest import mock

import attrs
import pytest
from aiohttp.abc import ResolveResult
from aiohttp.connector import TCPConnector
from aiohttp.test_utils import BaseTestServer, RawTestServer
from aiohttp.tracing import Trace
from aiohttp.web import Response as Response
from aiohttp.web_request import BaseRequest
from aiohttp.web_response import json_response
from yarl import URL

type _Response = (
    Callable[[], Response]
    | Callable[[BaseRequest], Awaitable[Response]]
    | dict[str, object]
    | list[object]
    | str
)


class AddRoutes(Protocol):
    def __call__(self, *routes: Route) -> None: ...


class NoRouteFoundError(AssertionError):
    pass


@attrs.frozen
class Route:
    url: URL = attrs.field(converter=URL)
    response: _Response
    method: str = attrs.field(converter=str.upper, default='GET')
    single_use: bool = False
    path_qs_pattern: re.Pattern[str] = attrs.field(init=False)

    def __attrs_post_init__(self):
        object.__setattr__(self, 'path_qs_pattern', re.compile(self.url.path_qs, re.IGNORECASE))

    def matches(self, request: BaseRequest):
        if self.method != request.method:
            return False
        elif self.url.host != re.escape(request.host):
            return False
        elif not self.path_qs_pattern.fullmatch(request.path_qs):
            return False
        return True


def prepare_mock_server_router():
    async def match_route(request: BaseRequest) -> Response:
        for i, route in enumerate(routes):
            if not route.matches(request):
                continue

            if route.single_use:
                del routes[i]

            match route.response:
                case Callable() as fn:
                    return await fn(request) if inspect.iscoroutinefunction(fn) else fn()  # pyright: ignore  # noqa: PGH003
                case str() as text:
                    return Response(body=text)
                case (dict() | list()) as json:
                    return json_response(data=json)
                case _:
                    raise ValueError(f'Unsupported response type: {type(route.response)}')

        else:
            raise NoRouteFoundError(f'No match found for <{request.method} {request.url}>')

    routes = list[Route]()

    def add_routes(*routes_: Route) -> None:
        routes.extend(routes_)

    return partial(RawTestServer, handler=match_route), add_routes


def patch_aiohttp(patcher: pytest.MonkeyPatch, mock_server: BaseTestServer):
    @partial(patcher.setattr, 'aiohttp.TCPConnector')
    class _(TCPConnector):
        async def _resolve_host(
            self, host: str, port: int, traces: Sequence[Trace] | None = None
        ) -> list[ResolveResult]:
            return [
                {
                    'hostname': host,
                    'host': '127.0.0.1',
                    'port': mock_server.port,  # pyright: ignore[reportReturnType]
                    'family': self._family,
                    'proto': 0,
                    'flags': 0,
                }
            ]

    patcher.setattr('aiohttp.ClientRequest.is_ssl', mock.Mock(return_value=False))



# pyright: reportUnusedFunction=false

from __future__ import annotations

import enum
from collections.abc import Hashable, Mapping, Sequence
from contextlib import contextmanager
from functools import partial
from inspect import signature
from typing import Any, Literal, Never, Self, TypedDict, overload

import attrs
import cattrs
from prompt_toolkit.application import Application
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import IsDone
from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    FormattedText,
    StyleAndTextTuples,
    to_formatted_text,
)
from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    HSplit,
    Layout,
    Window,
)
from prompt_toolkit.shortcuts import PromptSession
from prompt_toolkit.shortcuts.progress_bar import ProgressBar, ProgressBarCounter
from prompt_toolkit.shortcuts.progress_bar import formatters as pb_formatters
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator
from prompt_toolkit.widgets import Label

from .._utils.attrs import fauxfrozen
from .._utils.web import open_url


def make_attrs_field_validator(
    attrs_cls: type[attrs.AttrsInstance], attribute_name: str, converter: cattrs.Converter
):
    "Construct a one-off validator for an attrs field."

    fields = attrs.fields(attrs.resolve_types(attrs_cls))
    attribute = getattr(fields, attribute_name)

    field_wrapper = attrs.make_class(
        'field_wrapper',
        {
            attribute_name: attrs.field(
                **{
                    p: getattr(attribute, p)
                    for p in signature(attrs.field).parameters
                    if hasattr(attribute, p)
                }
            )
        },
    )

    class AttrsFieldValidator(Validator):
        def validate(self, document: Document):
            try:
                converter.structure({attribute_name: document.text}, field_wrapper)
            except Exception as exc:
                raise ValidationError(
                    0, '\n'.join(cattrs.transform_error(exc, format_exception=lambda e, _: str(e)))
                ) from exc

    return AttrsFieldValidator()


@fauxfrozen
class Choice[T]:
    label: str | StyleAndTextTuples
    value: T
    disabled: bool = False
    browser_url: str | None = None


@fauxfrozen
class _FauxPromptSession[T]:
    application: Application[T]

    def prompt(self) -> T:
        return self.application.run()

    async def prompt_async(self) -> T:
        return await self.application.run_async()


_style = Style(
    [
        ('indicator', 'fg:ansicyan'),
        ('question', 'bold'),
        ('skipped', 'fg:ansiyellow'),
        ('attention', 'fg:ansimagenta'),
    ]
)


class _Skip(enum.Enum):
    Skip = enum.auto()


SKIP = _Skip.Skip
_skip_choice = Choice([('underline', 's'), ('', 'kip')], SKIP)


def confirm(message: str) -> PromptSession[bool]:
    bindings = KeyBindings()

    result = None

    @bindings.add('y')
    @bindings.add('Y')
    def yes(event: KeyPressEvent) -> None:
        nonlocal result
        result = True
        event.app.exit(result=result)

    @bindings.add('n')
    @bindings.add('N')
    def no(event: KeyPressEvent) -> None:
        nonlocal result
        result = False
        event.app.exit(result=result)

    @bindings.add(Keys.Enter)  # Override <enter> from the prompt session's base bindings
    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent) -> None:
        "Disallow inserting other text."

    def get_messsage():
        tokens = [
            ('class:indicator', '?' if result is None else '✓' if result else '✗'),
            ('', ' '),
            ('class:question', message),
        ]
        if result is None:
            tokens += [('', '  (Y/n)')]
        return FormattedText(tokens)

    session = PromptSession[bool](
        get_messsage,
        key_bindings=bindings,
        style=_style,
    )
    return session


def path(
    message: str, *, only_directories: bool = False, validator: Validator | None = None
) -> PromptSession[str]:
    completer = PathCompleter(
        expanduser=True,
        only_directories=only_directories,
    )
    session = PromptSession[str](
        FormattedText(
            [('class:indicator', '>'), ('', ' '), ('class:question', message), ('', '  ')]
        ),
        completer=completer,
        style=_style,
        validator=validator,
    )
    return session


def text(message: str, *, is_password: bool = False) -> PromptSession[str]:
    session = PromptSession[str](
        FormattedText(
            [('class:indicator', '>'), ('', ' '), ('class:question', message), ('', '  ')]
        ),
        is_password=is_password,
        style=_style,
    )
    return session


password = partial(text, is_password=True)


def select_multiple[T](
    message: str,
    choices: Sequence[Choice[T]],
) -> _FauxPromptSession[list[T]]:
    bindings = KeyBindings()

    answered = False

    position = 0
    selected_indices = set[int]()

    @bindings.add(Keys.ControlC)
    @bindings.add(Keys.SIGINT)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Up)
    def previous_item(event: KeyPressEvent):
        nonlocal position
        position = max(0, position - 1)

    @bindings.add(Keys.Down)
    def next_item(event: KeyPressEvent):
        nonlocal position
        position = min(len(choices) - 1, position + 1)

    @bindings.add(' ')
    def select_item(event: KeyPressEvent):
        nonlocal selected_indices
        selected_indices = selected_indices ^ {position}

    @bindings.add('o')
    @bindings.add('O')
    def open_item_browser_url(event: KeyPressEvent):
        choice_at_position = choices[position]
        if choice_at_position.browser_url:
            open_url(choice_at_position.browser_url)

    @bindings.add(Keys.Enter)
    def submit(event: KeyPressEvent):
        nonlocal answered
        answered = True
        event.app.exit(result=[choices[i].value for i in selected_indices])

    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent):
        "Disallow inserting other text."

    def get_label_messsage():
        tokens = [
            ('class:indicator', '✓' if answered else '-'),
            ('', ' '),
            ('class:question', message),
        ]
        if not answered:
            tokens += [
                (
                    '',
                    '  (use arrow keys to move, <space> to select'
                    + (', <o> to open in browser' if any(c.browser_url for c in choices) else '')
                    + ' and <enter> to confirm)',
                )
            ]
        return FormattedText(tokens)

    def get_select_tokens():
        tokens = list[StyleAndTextTuples]()

        for i, choice in enumerate(choices):
            focussed = i == position
            if focussed:
                tokens += [('[SetCursorPosition]', '')]

            if choice.disabled:
                tokens += [('', ' ')]
            else:
                selected = i in selected_indices
                if selected:
                    tokens += [('class:checkbox-selected', '■')]
                else:
                    tokens += [('', '□')]

            tokens += [('', ' '), *to_formatted_text(choice.label), ('', '\n')]

        return tokens[:-1]

    app = Application[list[T]](
        key_bindings=bindings,
        layout=Layout(
            HSplit(
                [
                    Label(get_label_messsage),
                    ConditionalContainer(
                        Window(FormattedTextControl(get_select_tokens, focusable=True)),
                        filter=~IsDone(),
                    ),
                ],
            )
        ),
        style=_style,
    )
    return _FauxPromptSession(app)


@overload
def select_one[T](
    message: str,
    choices: Sequence[Choice[T]],
    *,
    can_skip: Literal[True],
    initial_value: T | None = None,
) -> _FauxPromptSession[T | Literal[_Skip.Skip]]: ...
@overload
def select_one[T](
    message: str,
    choices: Sequence[Choice[T]],
    *,
    can_skip: Literal[False] = False,
    initial_value: T | None = None,
) -> _FauxPromptSession[T]: ...


def select_one[T](
    message: str,
    choices: Sequence[Choice[T]],
    *,
    can_skip: bool = False,
    initial_value: T | None = None,
):
    bindings = KeyBindings()

    answered = False

    combined_choices = [*choices, _skip_choice] if can_skip else choices
    positions = [i for i, c in enumerate(combined_choices) if not c.disabled]
    position = next(
        p for p in positions if initial_value is None or choices[p].value == initial_value
    )

    @bindings.add(Keys.ControlC)
    @bindings.add(Keys.SIGINT)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Up)
    def previous_item(event: KeyPressEvent):
        nonlocal position
        try:
            position = positions[positions.index(position) - 1]
        except IndexError:
            position = positions[-1]

    @bindings.add(Keys.Down)
    def next_item(event: KeyPressEvent):
        nonlocal position
        try:
            position = positions[positions.index(position) + 1]
        except IndexError:
            position = positions[0]

    @bindings.add('o')
    @bindings.add('O')
    def open_item_browser_url(event: KeyPressEvent):
        choice_at_position = combined_choices[position]
        if choice_at_position.browser_url:
            open_url(choice_at_position.browser_url)

    @bindings.add(Keys.Enter)
    def submit(event: KeyPressEvent):
        nonlocal answered
        answered = True
        event.app.exit(result=combined_choices[position].value)

    if can_skip:

        @bindings.add('s')
        def skip_question(event: KeyPressEvent):
            nonlocal answered, position
            answered = True
            position = positions[-1]
            event.app.exit(result=SKIP)

    @bindings.add(Keys.Any)
    def _(event: KeyPressEvent):
        "Disallow inserting other text."

    def get_label_messsage():
        tokens = [
            (
                'class:indicator',
                '✓' if answered and combined_choices[position] is not _skip_choice else '-',
            ),
            ('', ' '),
            ('class:question', message),
        ]
        if answered and combined_choices[position] is _skip_choice:
            tokens += [('class:skipped', '  (skipped)')]
        elif not answered:
            tokens += [
                (
                    '',
                    '  (use arrow keys to move'
                    + (
                        ', <o> to open in browser'
                        if any(c.browser_url for c in combined_choices)
                        else ''
                    )
                    + ' and <enter> to select)',
                )
            ]
        else:
            tokens += [
                ('', '  ('),
                *to_formatted_text(combined_choices[position].label),
                ('', ')'),
            ]
        return FormattedText(tokens)

    def get_select_tokens():
        tokens = list[StyleAndTextTuples]()

        for i, choice in enumerate(combined_choices):
            if choice.disabled:
                tokens += [('', ' ')]
            else:
                focussed = i == position
                if focussed:
                    tokens += [('[SetCursorPosition]', ''), ('class:radio-selected', '●')]
                else:
                    tokens += [('', '○')]

            tokens += [('', ' '), *to_formatted_text(choice.label), ('', '\n')]

        return tokens[:-1]

    app = Application[Any](
        key_bindings=bindings,
        layout=Layout(
            HSplit(
                [
                    Label(get_label_messsage),
                    ConditionalContainer(
                        Window(FormattedTextControl(get_select_tokens, focusable=True)),
                        filter=~IsDone(),
                    ),
                ],
            )
        ),
        style=_style,
    )
    return _FauxPromptSession(app)


class _ProgressBarCounter(ProgressBarCounter[Never]):
    def __init__(
        self,
        *,
        progress_bar: ProgressBar,
        is_download: bool,
        label: AnyFormattedText = '',
        current: int = 0,
        total: int | None = None,
    ) -> None:
        super().__init__(progress_bar=progress_bar, label=label, total=total)
        self.is_download = is_download

    def with_progress(self, progress: int) -> Self:
        self.items_completed = progress
        return self


class _DownloadProgress(pb_formatters.Progress):
    template = HTML('<current>{current:>3.1f}</current>/<total>{total:>3.1f}</total>MB')

    def format(self, progress_bar: ProgressBar, progress: ProgressBarCounter[object], width: int):
        if isinstance(progress, _ProgressBarCounter) and not progress.is_download:
            return super().format(progress_bar, progress, width)

        return self.template.format(
            current=progress.items_completed / 2**20, total=(progress.total or 0) / 2**20
        )


class _ProgressStatus(TypedDict):
    label: str
    current: int
    total: int | None
    is_download: bool


@contextmanager
def make_progress_bar_group():
    "``ProgressBar`` with download progress expressed in megabytes."

    progress_bar = ProgressBar(
        formatters=[
            pb_formatters.Label(),
            pb_formatters.Text(' '),
            pb_formatters.Percentage(),
            pb_formatters.Text(' '),
            pb_formatters.Bar(sym_a='#', sym_b='#', sym_c='.'),
            pb_formatters.Text(' '),
            _DownloadProgress(),
        ],
    )

    progress_bars = dict[Hashable, _ProgressBarCounter]()

    def update_progress(progress_statuses: Mapping[Hashable, _ProgressStatus]):
        nonlocal progress_bars

        progress_bars = {
            k: (e or _ProgressBarCounter(progress_bar=progress_bar, **p)).with_progress(
                p['current']
            )
            for (k, p), e in zip(
                progress_statuses.items(), (progress_bars.get(k) for k in progress_statuses)
            )
        }

        progress_bar.counters = list(progress_bars.values())
        progress_bar.invalidate()

    with progress_bar:
        try:
            yield update_progress

        finally:
            update_progress({})



from __future__ import annotations

from collections.abc import Mapping
from enum import Enum

import click
import click.types

from .._utils.iteration import bucketise


class EnumValueChoiceParam[EnumT: Enum](click.Choice[EnumT]):
    def __init__(
        self,
        choice_enum: type[EnumT],
    ):
        super().__init__(choices=list(choice_enum))
        self.__choice_enum = choice_enum

    def normalize_choice(self, choice: EnumT, ctx: click.Context | None) -> str:
        return super().normalize_choice(self.__choice_enum(choice), ctx)


class ManyOptionalChoiceValueParam[ParamT](click.types.CompositeParamType):
    name = 'optional-choice-value'

    def __init__(
        self,
        choice_param: click.Choice[ParamT],
        *,
        value_types: Mapping[str, type] = {},
    ):
        super().__init__()
        self.__choice_param = choice_param
        self.__value_types = {k: click.types.convert_type(v) for k, v in value_types.items()}

    @property
    def arity(self):
        return -1

    def convert(
        self, value: tuple[str, ...], param: click.Parameter | None, ctx: click.Context | None
    ):
        def do_convert(raw_entries: tuple[str, ...]):
            converter = self.__choice_param
            value_types = self.__value_types
            for raw_entry in raw_entries:
                key, sep, value = raw_entry.partition('=')
                value_type = value_types.get(key)
                value = value if sep else None
                yield (
                    converter.convert(key, param, ctx),
                    value_type.convert(value, param, ctx) if value_type and value else value,
                )

        return dict(do_convert(value))

    def get_metavar(self, param: click.Parameter, ctx: click.Context):
        return f'{{{",".join(map(str, self.__choice_param.choices))}}}[=VALUE]'


class SectionedHelpGroup(click.Group):
    group_class = type

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter):
        command_sections = bucketise(
            ((s, c) for s, c in self.commands.items() if not c.hidden),
            key=lambda c: 'Command groups' if isinstance(c[1], click.Group) else 'Commands',
        )
        if command_sections:
            for section_name, commands in command_sections.items():
                with formatter.section(section_name):
                    limit = formatter.width - 6 - max(len(s) for s, _ in commands)
                    formatter.write_dl(
                        [(s, c.get_short_help_str(limit)) for s, c in commands],
                    )



from __future__ import annotations

import datetime as dt
from collections.abc import Iterator, Mapping, Sequence
from itertools import chain, product
from operator import itemgetter
from pathlib import Path
from typing import Any, Literal, NotRequired, Protocol, TypedDict

import cattrs
from yarl import URL

from instawow import http, http_ctx
from instawow._logging import logger
from instawow._utils.aio import gather, run_in_thread
from instawow._utils.attrs import fauxfrozen
from instawow._utils.iteration import bucketise, uniq
from instawow._utils.perf import time_op
from instawow.progress_reporting import make_download_progress, make_incrementing_progress_tracker
from instawow.wow_installations import (
    extract_installation_dir_from_addon_dir,
    extract_installation_version_from_addon_dir,
)

from ._utils import get_checksum
from .config import PluginConfig

_saved_vars_cache_name = '_saved_vars_v1'

_api_base_url = URL('https://data.wago.io/api')

_aura_converter = cattrs.Converter()
_aura_converter.register_structure_hook(URL, lambda v, _: URL(v))
_aura_converter.register_unstructure_hook(URL, str)


class _WagoApiResponse(TypedDict):
    _id: str
    'Alphanumeric ID'
    name: str
    'User-facing name'
    slug: str
    'Slug if it has one; otherwise same as ``_id``'
    url: str
    created: str
    'ISO datetime'
    modified: str
    'ISO datetime'
    game: str
    'Xpac name, e.g. "bfa"'
    username: NotRequired[str]
    'Author username'
    version: int
    'Version counter, incremented with every update'
    versionString: str
    'Semver auto-generated from ``version`` - for presentation only'
    changelog: _WagoApiResponse_Changelog
    forkOf: NotRequired[str]
    'Only present on forks'
    regionType: NotRequired[str]
    'Only present on WAs'


class _WagoApiResponse_Changelog(TypedDict):
    format: NotRequired[Literal['bbcode', 'markdown']]
    text: NotRequired[str]


@fauxfrozen(kw_only=True)
class _Aura:
    id: str
    uid: str
    parent: str | None = None
    url: URL
    version: int


type _AuraGroup = dict[str, list[_Aura]]


class _AuraAddon(Protocol):  # pragma: no cover
    name: str
    api_endpoint: str

    def extract_auras(self, lua_table: Any) -> _AuraGroup: ...


@object.__new__
class _WeakAuras(_AuraAddon):
    name = 'WeakAuras'
    api_endpoint = 'weakauras'

    def extract_auras(self, lua_table: Any):
        raw_auras = lua_table['WeakAurasSaved']['displays']
        prepared_auras = (t for t in raw_auras.values() if t.get('url'))
        auras = _aura_converter.structure(prepared_auras, list[_Aura])
        sorted_auras = sorted(
            filter(lambda a: not a.parent and a.url.host == 'wago.io', auras), key=lambda a: a.id
        )
        return bucketise(sorted_auras, key=lambda a: a.url.parts[1])


@object.__new__
class _Plateroos(_AuraAddon):
    name = 'Plater'
    api_endpoint = 'plater'

    def extract_auras(self, lua_table: Any):
        raw_auras = lua_table['PlaterDB']['profiles']
        prepared_auras = (
            {**t, 'id': t['Name'], 'uid': ''}
            for n, p in raw_auras.items()
            for t in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                p.get('script_data') or (),
                p.get('hook_data') or (),
            )
            if t.get('url')
        )
        auras = _aura_converter.structure(prepared_auras, list[_Aura])
        sorted_auras = sorted(
            filter(lambda a: not a.parent and a.url.host == 'wago.io', auras), key=lambda a: a.id
        )
        return {a.url.parts[1]: [a] for a in sorted_auras}


@fauxfrozen
class _BuildPaths:
    archive: Path
    changelog: Path
    version: Path


async def _get_remote_auras(
    plugin_config: PluginConfig, aura_addon: _AuraAddon, aura_ids: list[str]
):
    if not aura_ids:
        return ()

    access_token = plugin_config.access_tokens.wago
    request_headers = {'api-key': access_token} if access_token else None

    async with http_ctx.web_client().post(
        _api_base_url / 'check' / aura_addon.api_endpoint,
        expire_after=dt.timedelta(minutes=30),
        headers=request_headers,
        json={'ids': aura_ids},
        trace_request_ctx={'progress': make_download_progress(label='Fetching aura metadata')},
    ) as response:
        if response.status == 404:
            return ()

        response.raise_for_status()

        metadata: list[_WagoApiResponse] = sorted(
            await response.json(),
            key=itemgetter('slug'),
        )

    async def fetch_wago_import_string(remote_aura: _WagoApiResponse):
        async with http_ctx.web_client().get(
            (_api_base_url / 'raw' / 'encoded')
            .with_query(id=remote_aura['_id'])
            .with_fragment(str(remote_aura['version'])),
            expire_after=http.CACHE_INDEFINITELY,
            headers=request_headers,
            raise_for_status=True,
        ) as response:
            return await response.text()

    track_import_string_progress = make_incrementing_progress_tracker(
        len(metadata), f'Fetching import strings for {aura_addon.name}'
    )
    import_strings = await gather(
        track_import_string_progress(fetch_wago_import_string(r)) for r in metadata
    )

    return list(zip(metadata, import_strings))


def _generate_addon(
    plugin_config: PluginConfig,
    aura_groups: Mapping[_AuraAddon, Sequence[tuple[_WagoApiResponse, str]]],
):
    from zipfile import ZipFile, ZipInfo

    build_paths = get_build_paths(
        plugin_config.ensure_dirs(),
    )

    with ZipFile(build_paths.archive, 'w') as file:

        def write_file(filename: str, content: str):
            # Not using a plain string as the first argument to ``writestr``
            # 'cause the timestamp would be set to the current time
            # which would render the build unreproducible.
            file.writestr(ZipInfo(filename=f'WeakAurasCompanion/{filename}'), content)
            return content

        data_output = write_file(
            'data.lua',
            f"""\
-- file generated automatically
WeakAurasCompanionData = {{
{
                '\n'.join(
                    f'''\
    {addon.name} = {{
        slugs = {{
{
                        '\n'.join(
                            f"""\
            [ [=[{metadata['slug']}]=] ] = {{
                name = [=[{metadata['name']}]=],
                author = [=[{metadata.get('username', '__unknown__')}]=],
                encoded = [=[{import_string}]=],
                wagoVersion = [=[{metadata['version']}]=],
                wagoSemver = [=[{metadata['version']}]=],
                versionNote = [=[{metadata['changelog'].get('text', '')}]=],
                source = [=[Wago]=],
            }},"""
                            for metadata, import_string in g
                        )
                    }
        }},
        stash = {{
        }},
        stopmotionFiles = {{
        }},
    }},'''
                    for addon, g in aura_groups.items()
                )
            }
}}
""",
        )

        init_output = write_file(
            'init.lua',
            """\
-- file generated automatically
local loadedFrame = CreateFrame("FRAME")
loadedFrame:RegisterEvent("ADDON_LOADED")
loadedFrame:SetScript("OnEvent", function(_, _, addonName)
  if addonName == "WeakAurasCompanion" then
    if WeakAuras and WeakAuras.AddCompanionData and WeakAurasCompanionData then
      local WeakAurasData = WeakAurasCompanionData.WeakAuras
      if WeakAurasData then
        WeakAuras.AddCompanionData(WeakAurasData)
        WeakAuras.StopMotion.texture_types["WeakAuras Companion"] = WeakAuras.StopMotion.texture_types["WeakAuras Companion"] or {}
        local CompanionTextures = WeakAuras.StopMotion.texture_types["WeakAuras Companion"]
        for fileName, name in pairs(WeakAurasData.stopmotionFiles) do
          CompanionTextures["Interface\\\\AddOns\\\\WeakAurasCompanion\\\\animations\\\\" .. fileName] = name
        end
      end
    end

    if Plater and Plater.AddCompanionData and WeakAurasCompanionData and WeakAurasCompanionData.Plater then
      Plater.AddCompanionData(WeakAurasCompanionData.Plater)
    end
  end
end)
""",
        )

        interface_version = (
            extract_installation_version_from_addon_dir(plugin_config.profile_config.addon_dir)
            or 0
        )
        addon_version = get_checksum(data_output, init_output, interface_version)[:7]

        write_file(
            'WeakAurasCompanion.toc',
            f"""\
## Interface: {interface_version}
## Title: WeakAuras Companion
## Author: The WeakAuras Team
## X-Generator: instawow
## Version: {addon_version}
## IconTexture: Interface\\AddOns\\WeakAuras\\Media\\Textures\\icon.blp
## Notes: Keep your WeakAuras updated!
## DefaultState: Enabled
## LoadOnDemand: 0
## Dependencies: WeakAuras
## OptionalDeps: {','.join(a.name for a in aura_groups if a.name != 'WeakAuras')}

data.lua
init.lua
""",
        )

    build_paths.changelog.write_text(
        '\n\n'.join(
            f"""\
## {m['name']} v{m['version']} ({m['url']})

{m['changelog'].get('text') or 'n/a'}
"""
            for g in aura_groups.values()
            for m, _ in g
        )
        or 'n/a',
        encoding='utf-8',
    )

    build_paths.version.write_text(
        addon_version,
        encoding='utf-8',
    )

    return build_paths


async def build_addon(plugin_config: PluginConfig) -> _BuildPaths:
    installed_auras = await run_in_thread(list)(extract_installed_auras(plugin_config))
    grouped_aura_ids = {
        a: uniq(k for *_, g in v for k in g)
        for a, v in bucketise(installed_auras, key=itemgetter(1)).items()
    }
    remote_auras = await gather(
        _get_remote_auras(plugin_config, *v) for v in grouped_aura_ids.items()
    )
    return await run_in_thread(_generate_addon)(
        plugin_config, dict(zip(grouped_aura_ids, remote_auras))
    )


def extract_installed_auras(
    plugin_config: PluginConfig,
) -> Iterator[tuple[str, _AuraAddon, _AuraGroup]]:
    import diskcache

    installation_dir = extract_installation_dir_from_addon_dir(
        plugin_config.profile_config.addon_dir
    )
    if not installation_dir:
        raise ValueError(
            f'Cannot determine installation folder from {plugin_config.profile_config.addon_dir}'
        )

    with diskcache.Cache(
        plugin_config.dirs.cache / _saved_vars_cache_name,
    ) as cache:
        for account_sv_path, addon in product(
            installation_dir.glob('WTF/Account/*/SavedVariables'), (_WeakAuras, _Plateroos)
        ):
            sv_path = (account_sv_path / addon.name).with_suffix('.lua')
            if not sv_path.exists():
                continue

            content = sv_path.read_text(encoding='utf-8-sig', errors='replace')
            checksum_key = f'saved_vars_checksum__{sv_path}'
            checksum = get_checksum(content)
            auras_key = f'saved_vars_auras__{sv_path}'

            if cache.get(checksum_key) == checksum:
                logger.info(f'Loading auras from cache for {sv_path!r}')
                auras = cache[auras_key]
            else:
                with time_op(
                    lambda t: logger.debug(
                        f'Extracted {addon.name!r} auras in {t:.3f}s from {sv_path!r}'  # noqa: B023
                    )
                ):
                    from . import _custom_slpp

                    auras = addon.extract_auras(_custom_slpp.loads(f'{{ {content} }}'))

                with cache.transact():
                    cache[checksum_key] = checksum
                    cache[auras_key] = auras

            yield account_sv_path.parent.name, addon, auras


def get_build_paths(plugin_config: PluginConfig) -> _BuildPaths:
    return _BuildPaths(
        plugin_config.profile_cache_path / 'WeakAurasCompanion.zip',
        plugin_config.profile_cache_path / 'CHANGELOG.md',
        plugin_config.profile_cache_path / 'VERSION',
    )



from __future__ import annotations

import sys
from enum import Enum
from pathlib import Path
from unittest import mock

import pytest

from instawow.config import ProfileConfig
from instawow.wow_installations import (
    Flavour,
    FlavourVersions,
    Track,
    extract_installation_version_from_addon_dir,
    find_installations,
    get_compatible_flavours,
    infer_flavour_from_addon_dir,
    to_flavourful_enum,
)


def test_can_convert_between_flavour_keyed_enum_and_flavour():
    class Foo(Enum):
        Retail = 1

    assert to_flavourful_enum(Foo.Retail, Flavour) is Flavour.Retail
    assert to_flavourful_enum(Flavour.Retail, Foo) is Foo.Retail


@pytest.mark.parametrize(
    ('track', 'affine', 'expected_flavours'),
    [
        (Track.Retail, False, (Flavour.Retail,)),
        (Track.Retail, True, (Flavour.Retail, None)),
        (Track.VanillaClassic, False, (Flavour.VanillaClassic,)),
        (Track.VanillaClassic, True, (Flavour.VanillaClassic, None)),
        (Track.Classic, False, (Flavour.MistsClassic, Flavour.CataClassic)),
        (Track.Classic, True, (Flavour.MistsClassic, None)),
    ],
)
def test_compatible_flavours_vary_by_track_and_affinity(
    track: Track, affine: bool, expected_flavours: tuple[Flavour | None, ...]
):
    assert get_compatible_flavours(track, affine) == expected_flavours


def test_can_extract_flavour_from_version_number():
    assert FlavourVersions.from_version_number(9_50_00) is FlavourVersions.Retail
    assert FlavourVersions.from_version_number(4_04_00) is FlavourVersions.CataClassic
    assert FlavourVersions.from_version_number(5_05_00) is FlavourVersions.MistsClassic
    assert FlavourVersions.from_version_number(1_23_00) is FlavourVersions.VanillaClassic


def test_can_extract_flavour_from_version_string():
    assert FlavourVersions.from_version_string('9.50.0') is FlavourVersions.Retail
    assert FlavourVersions.from_version_string('4.4.0') is FlavourVersions.CataClassic
    assert FlavourVersions.from_version_string('5.5.0') is FlavourVersions.MistsClassic
    assert FlavourVersions.from_version_string('1.23.0') is FlavourVersions.VanillaClassic


def test_can_extract_flavour_from_partial_version_string():
    assert FlavourVersions.from_version_string('9.2') is FlavourVersions.Retail
    assert FlavourVersions.from_version_string('4.4') is FlavourVersions.CataClassic
    assert FlavourVersions.from_version_string('5.5') is FlavourVersions.MistsClassic
    assert FlavourVersions.from_version_string('3') is FlavourVersions.Retail


@pytest.mark.parametrize(
    ('path', 'flavour'),
    [
        (
            'wowzerz/_classic_/Interface/AddOns',
            Flavour.MistsClassic,
        ),
        (
            '/foo/bar/_classic_ptr_/Interface/AddOns',
            Flavour.MistsClassic,
        ),
        (
            '_classic_era_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            '_classic_era_ptr_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            'wowzerz/_retail_/Interface/AddOns',
            Flavour.Retail,
        ),
        (
            'anything goes',
            None,
        ),
    ],
)
def test_can_infer_flavour_from_addon_dir(
    path: str,
    flavour: Flavour | None,
):
    assert infer_flavour_from_addon_dir(path) is flavour


@pytest.mark.skipif(sys.platform != 'darwin', reason='Only supported on Mac')
def test_can_find_mac_installations(
    monkeypatch: pytest.MonkeyPatch,
):
    with monkeypatch.context() as patcher:
        patcher.setattr('subprocess.check_output', mock.Mock(return_value=''))
        assert not list(find_installations())

    with monkeypatch.context() as patcher:
        app_bundle_paths = {
            Path('/Applications/World of Warcraft/_retail_/World of Warcraft.app'): {
                'code': 'wow',
                'flavour': Flavour.Retail,
            },
            Path('/Applications/World of Warcraft/_classic_/World of Warcraft Classic.app'): {
                'code': 'wow_classic',
                'flavour': Flavour.MistsClassic,
            },
        }

        patcher.setattr(
            'subprocess.check_output',
            mock.Mock(return_value='\n'.join(map(str, app_bundle_paths))),
        )
        assert list(find_installations()) == [(p.parent, f) for p, f in app_bundle_paths.items()]


def test_installation_version_extraction_from_addon_dir(
    iw_profile_config: ProfileConfig,
):
    iw_profile_config.addon_dir.parents[2].joinpath('.build.info').write_bytes(b"""\
Version!STRING:0|Product!STRING:0
10.9.8.7|wow
""")
    assert extract_installation_version_from_addon_dir(iw_profile_config.addon_dir) == 100908



from __future__ import annotations

import datetime as dt

import pytest

from instawow import config_ctx, pkg_management
from instawow.catalogue.search import search
from instawow.definitions import Defn
from instawow.results import PkgInstalled
from instawow.wow_installations import Track

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


async def test_basic_search():
    limit = 5
    results = await search('masque', limit=limit)
    assert len(results) <= 5
    assert {('curse', 'masque'), ('wowi', '12097')} <= {
        (e.source, e.slug or e.id) for e in results
    }


@pytest.mark.parametrize(
    'iw_profile_config_values',
    Track,
    indirect=True,
)
async def test_search_flavour_filtering():
    results = await search('atlas loot classic', limit=10)
    has_atlas = ('curse', 'atlaslootclassic') in {(e.source, e.slug or e.id) for e in results}
    assert has_atlas == (config_ctx.config().track is Track.VanillaClassic)


async def test_search_source_filtering():
    results = await search('masque', limit=5, sources={'curse'})
    assert all(e.source == 'curse' for e in results)
    assert ('curse', 'masque') in {(e.source, e.slug) for e in results}


async def test_search_date_filtering():
    start_date = dt.datetime.now(tz=dt.UTC) - dt.timedelta(days=365)
    results = await search('masque', limit=5, start_date=start_date)
    assert all(e.last_updated > start_date for e in results)


async def test_search_unknown_source():
    with pytest.raises(ValueError, match='Unknown source'):
        await search('masque', limit=5, sources={'foo'})


async def test_search_filter_installed():
    results = await search('masque', limit=5, filter_installed='include_only')
    assert not results

    defn = Defn('curse', 'masque')

    install_result = (await pkg_management.install([defn], replace_folders=False))[defn]
    assert type(install_result) is PkgInstalled

    results = await search('masque', limit=5)
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} <= {
        (e.source, e.slug) for e in results
    }

    results = await search('masque', limit=5, filter_installed='include_only')
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} & {
        (e.source, e.slug) for e in results
    } == {('curse', 'masque')}

    results = await search('masque', limit=5, filter_installed='exclude')
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'sfx-wow/masque')}

    results = await search('masque', limit=5, filter_installed='exclude_from_all_sources')
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} & {
        (e.source, e.slug) for e in results
    } == set()


async def test_search_prefer_known_source():
    results = await search('masque', limit=5, prefer_source=None)
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} <= {
        (e.source, e.slug) for e in results
    }

    results = await search('masque', limit=5, prefer_source='github')
    assert {('curse', 'masque'), ('github', 'sfx-wow/masque')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'sfx-wow/masque')}


async def test_search_prefer_unknown_source():
    with pytest.raises(ValueError, match='Unknown preferred source: foo'):
        await search('masque', limit=5, prefer_source='foo')



from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import Any

import cattrs
import pytest

from instawow.config import GlobalConfig, ProfileConfig, UninitialisedConfigError, make_plugin_dirs


def test_top_level_env_vars_take_precedence(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setenv('INSTAWOW_AUTO_UPDATE_CHECK', '1')
    global_config = GlobalConfig.from_values({'auto_update_check': False}, env=True)
    assert global_config.auto_update_check is True


def test_nested_env_vars_take_precedence(
    monkeypatch: pytest.MonkeyPatch,
):
    override = 'utter nonsense'
    monkeypatch.setenv('INSTAWOW_ACCESS_TOKENS_GITHUB', override)
    global_config = GlobalConfig.from_values({'access_tokens': {'github': 'true story'}}, env=True)
    assert global_config.access_tokens.github == override


def test_nested_env_var_objs_take_precedence(
    monkeypatch: pytest.MonkeyPatch,
):
    override = 'utter nonsense'
    monkeypatch.setenv('INSTAWOW_ACCESS_TOKENS', json.dumps({'github': override}))
    monkeypatch.setenv('INSTAWOW_ACCESS_TOKENS_GITHUB', 'true story')
    global_config = GlobalConfig.from_values(env=True)
    assert global_config.access_tokens.github == override


def test_nested_env_vars_loaded_when_values_absent(
    monkeypatch: pytest.MonkeyPatch,
):
    override = 'utter nonsense'
    monkeypatch.setenv('INSTAWOW_ACCESS_TOKENS_GITHUB', override)
    global_config = GlobalConfig.from_values(env=True)
    assert global_config.access_tokens.github == override


def test_read_profile_from_nonexistent_config_dir_raises():
    global_config = GlobalConfig()
    with pytest.raises(UninitialisedConfigError):
        ProfileConfig.read(global_config, '__default__')


def test_init_with_nonexistent_addon_dir_raises(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig().write()
    with pytest.raises(cattrs.ClassValidationError) as exc_info:
        ProfileConfig.from_values(
            iw_profile_config_values | {'global_config': global_config, 'addon_dir': '#@$foo'}
        )
    assert exc_info.group_contains(ValueError, match='not a writable directory')


def test_default_config_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('INSTAWOW_HOME')
    monkeypatch.delenv('XDG_CONFIG_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        assert GlobalConfig().dirs.config == Path.home() / '.config/instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')

        assert GlobalConfig().dirs.config == Path.home() / 'Library/Application Support/instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'win32')

        patcher.delenv('APPDATA', False)
        assert GlobalConfig().dirs.config == Path.home() / '.config' / 'instawow'

        patcher.setenv('APPDATA', '/foo')
        assert GlobalConfig().dirs.config == Path('/foo/instawow').resolve()


def test_default_state_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('INSTAWOW_HOME')
    monkeypatch.delenv('XDG_STATE_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        assert GlobalConfig().dirs.state == Path.home() / '.local' / 'state' / 'instawow'

    for platform in {'darwin', 'win32'}:
        with monkeypatch.context() as patcher:
            patcher.setattr(sys, 'platform', platform)

            global_config = GlobalConfig()
            assert global_config.dirs.config == global_config.dirs.state


def test_default_cache_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('INSTAWOW_HOME')
    monkeypatch.delenv('XDG_CACHE_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        assert GlobalConfig().dirs.cache == Path.home() / '.cache' / 'instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')

        assert GlobalConfig().dirs.cache == Path.home() / 'Library/Caches/instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'win32')

        patcher.delenv('LOCALAPPDATA', False)
        assert GlobalConfig().dirs.cache == Path.home() / '.cache' / 'instawow'

        patcher.setenv('LOCALAPPDATA', '/foo')
        assert GlobalConfig().dirs.cache == Path('/foo/instawow').resolve()


def test_home_env_var_respected(
    iw_home: Path,
):
    global_config = GlobalConfig()
    assert global_config.dirs.config == iw_home / 'config'
    assert global_config.dirs.state == iw_home / 'state'
    assert global_config.dirs.cache == iw_home / 'cache'


def test_xdg_env_vars_respected(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    monkeypatch.delenv('INSTAWOW_HOME')

    x_homes = {c: tmp_path / f'iw_{c}_home' for c in ('config', 'state', 'cache')}
    for dir_name, path in x_homes.items():
        monkeypatch.setenv(f'XDG_{dir_name.upper()}_HOME', str(path))

    global_config = GlobalConfig()
    assert global_config.dirs.config == x_homes['config'] / 'instawow'
    assert global_config.dirs.state == x_homes['state'] / 'instawow'
    assert global_config.dirs.cache == x_homes['cache'] / 'instawow'


def test_plugin_dir_names():
    global_config_dirs = GlobalConfig().dirs
    plugin_dirs = make_plugin_dirs('foo')
    assert global_config_dirs.config / 'plugins' / 'foo' == plugin_dirs.config
    assert global_config_dirs.cache / 'plugins' / 'foo' == plugin_dirs.cache
    assert global_config_dirs.state / 'plugins' / 'foo' == plugin_dirs.state


def test_access_tokens_file_takes_precedence_over_config():
    global_config = GlobalConfig.read().write()
    assert global_config.access_tokens.cfcore is None
    global_config.dirs.config.joinpath('config.access_tokens.json').write_text(
        json.dumps({'cfcore': 'abc'}), encoding='utf-8'
    )
    assert GlobalConfig.read().access_tokens.cfcore == 'abc'


def test_can_list_profiles(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.read()
    assert set(ProfileConfig.iter_profiles(global_config)) == set()

    ProfileConfig.from_values(iw_profile_config_values | {'global_config': global_config}).write()
    ProfileConfig.from_values(
        iw_profile_config_values | {'global_config': global_config, 'profile': 'foo'}
    ).write()
    assert set(ProfileConfig.iter_profiles(global_config)) == {'__default__', 'foo'}


def test_can_list_installations(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.read()
    assert set(ProfileConfig.iter_profile_installations(global_config)) == set()

    ProfileConfig.from_values(iw_profile_config_values | {'global_config': global_config}).write()
    assert set(ProfileConfig.iter_profile_installations(global_config)) == {
        iw_profile_config_values['_installation_dir']
    }


def test_profile_dirs_populated(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig()
    profile_config = ProfileConfig.from_values(
        iw_profile_config_values | {'global_config': global_config}
    ).write()
    assert {i.name for i in profile_config.config_path.iterdir()}.issuperset({'config.json'})


def test_can_delete_profile(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig().write()
    profile_config = ProfileConfig.from_values(
        iw_profile_config_values | {'global_config': global_config}
    ).write()
    assert profile_config.config_path.exists()
    profile_config.delete()
    assert not profile_config.config_path.exists()


def test_profiled_name_validated(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig()

    with pytest.raises(cattrs.ClassValidationError) as exc_info:
        ProfileConfig.from_values(
            iw_profile_config_values | {'global_config': global_config, 'profile': ''}
        )

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == ('Value must have a minimum length of 1',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class ProfileConfig @ attribute profile'
    assert type(note) is cattrs.AttributeValidationNote
    assert note.name == 'profile'


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='chmod has no effect on Windows',
)
def test_addon_dir_validated(
    tmp_path: Path,
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig()

    non_writeable_dir = tmp_path / 'non-writeable-dir'
    non_writeable_dir.mkdir(0o400)

    with pytest.raises(cattrs.ClassValidationError) as exc_info:
        ProfileConfig.from_values(
            iw_profile_config_values
            | {'global_config': global_config, 'addon_dir': non_writeable_dir}
        )

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == (f'"{non_writeable_dir}" is not a writable directory',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class ProfileConfig @ attribute addon_dir'
    assert type(note) is cattrs.AttributeValidationNote
    assert note.name == 'addon_dir'



from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from enum import Enum, StrEnum
from functools import cache, partial
from pathlib import Path
from typing import Self, TypedDict

from ._utils.iteration import fill


class Track(StrEnum):
    Retail = 'retail'
    VanillaClassic = 'vanilla_classic'
    MistsClassic = 'classic'
    Classic = MistsClassic


class Flavour(StrEnum):
    Retail = 'retail'
    VanillaClassic = 'vanilla_classic'
    TbcClassic = 'tbc_classic'
    WrathClassic = 'wrath_classic'
    CataClassic = 'cata_classic'
    MistsClassic = 'mists_classic'


class FlavourVersions(Enum):
    Retail = (
        range(1_00_00, 1_13_00),
        range(2_00_00, 2_05_00),
        range(3_00_00, 3_04_00),
        range(4_00_00, 4_04_00),
        range(5_00_00, 5_05_00),
        range(6_00_00, 12_00_00),
    )
    VanillaClassic = (range(1_13_00, 2_00_00),)
    TbcClassic = (range(2_05_00, 3_00_00),)
    WrathClassic = (range(3_04_00, 4_00_00),)
    CataClassic = (range(4_04_00, 5_00_00),)
    MistsClassic = (range(5_05_00, 6_00_00),)

    @classmethod
    def from_version_number(cls, version: int) -> Self | None:
        return next((f for f in cls for r in f.value if version in r), None)

    @classmethod
    def from_version_string(cls, version: str) -> Self | None:
        return cls.from_version_number(_parse_version_string(version))


class FlavourTocSuffixes(Enum):
    # https://github.com/Stanzilla/WoWUIBugs/issues/68#issuecomment-830351390
    # https://warcraft.wiki.gg/wiki/TOC_format#Multiple_client_flavors
    Retail = ('Mainline',)
    VanillaClassic = ('Vanilla', 'Classic')
    TbcClassic = ('TBC', 'BCC', 'Classic')
    WrathClassic = ('Wrath', 'WOTLKC', 'Classic')
    CataClassic = ('Cata', 'Classic')
    MistsClassic = ('Mists', 'Classic')


def get_compatible_flavours(
    track: Track, affine: bool | None = None
) -> tuple[Flavour, ...] | tuple[*tuple[Flavour, ...], None]:
    match (track, affine):
        case (_, True):
            return (to_flavour(track), None)
        case (Track.Classic, _):
            return (Flavour.MistsClassic, Flavour.CataClassic)
        case _:
            return (to_flavour(track),)


def to_flavourful_enum[TargetEnumT: Enum](
    source_enum: Enum, target_enum_type: type[TargetEnumT]
) -> TargetEnumT:
    return target_enum_type[source_enum.name]


to_flavour = partial(to_flavourful_enum, target_enum_type=Flavour)
to_flavour_versions = partial(to_flavourful_enum, target_enum_type=FlavourVersions)
to_flavour_toc_suffixes = partial(to_flavourful_enum, target_enum_type=FlavourTocSuffixes)


class _Product(TypedDict):
    code: str
    flavour: Flavour


_DELECTABLE_DIR_NAMES: dict[str, _Product] = {
    '_retail_': {
        'code': 'wow',
        'flavour': Flavour.Retail,
    },
    '_ptr_': {
        'code': 'wowt',
        'flavour': Flavour.Retail,
    },
    '_xptr_': {
        'code': 'wowxptr',
        'flavour': Flavour.Retail,
    },
    '_beta_': {
        'code': 'wow_beta',
        'flavour': Flavour.Retail,
    },
    '_classic_era_': {
        'code': 'wow_classic_era',
        'flavour': Flavour.VanillaClassic,
    },
    '_classic_era_ptr_': {
        'code': 'wow_classic_era_ptr',
        'flavour': Flavour.VanillaClassic,
    },
    # '_classic_era_beta_': {
    #     'code': 'wow_classic_era_beta',
    #     'flavour': Flavour.VanillaClassic,
    # },
    '_classic_': {
        'code': 'wow_classic',
        'flavour': Flavour.MistsClassic,
    },
    '_classic_ptr_': {
        'code': 'wow_classic_ptr',
        'flavour': Flavour.MistsClassic,
    },
    '_classic_beta_': {
        'code': 'wow_classic_beta',
        'flavour': Flavour.MistsClassic,
    },
}

_ADDON_DIR_PARTS = ('Interface', 'AddOns')
_NORMALISED_ADDON_DIR_PARTS = tuple(map(str.casefold, _ADDON_DIR_PARTS))


def _find_mac_installations():
    import subprocess

    try:
        possible_installations = subprocess.check_output(
            [
                'mdfind',
                "kMDItemCFBundleIdentifier == 'com.blizzard.worldofwarcraft'",
            ],
            text=True,
            timeout=3,
        ).splitlines()
    except subprocess.TimeoutExpired:
        pass
    else:
        for match in possible_installations:
            installation_path = Path(match).parent
            yield (installation_path, _DELECTABLE_DIR_NAMES.get(installation_path.name))


def find_installations() -> Iterator[tuple[Path, _Product | None]]:
    if sys.platform == 'darwin':
        yield from _find_mac_installations()


def _parse_version_string(version_string: str) -> int:
    major, minor, patch = fill(map(int, version_string.split('.')), 0, 3)
    return major * 1_00_00 + minor * 1_00 + patch


def _read_info(info_path: Path):
    with open(info_path, encoding='utf-8') as info:
        header = next(info)
        fields = tuple(f[: f.find('!')] for f in header.rstrip().split('|'))
        return [dict(zip(fields, e.rstrip().split('|'))) for e in info if not e.startswith('#')]


@cache
def _extract_installed_versions_from_build_info(outer_installation_path: Path):
    return {
        e['Product']: _parse_version_string(e['Version'])
        for e in _read_info(outer_installation_path / '.build.info')
    }


@cache
def _extract_installed_version_from_config_wtf(installation_path: Path):
    with open(installation_path.joinpath('WTF', 'Config.wtf'), 'rb') as config_wtf:
        version_prefix = b'SET lastAddonVersion'
        for line in config_wtf:
            if line.startswith(version_prefix):
                byte_version = line.replace(version_prefix, b'').strip().strip(b'"')
                if byte_version:
                    return int(byte_version)
                break


def extract_installation_dir_from_addon_dir(path_like: os.PathLike[str] | str) -> Path | None:
    path = Path(path_like)
    tail = tuple(map(str.casefold, path.parts[-3:]))
    if len(tail) == 3 and tail[1:] == _NORMALISED_ADDON_DIR_PARTS:
        return path.parents[1]


def extract_installation_version_from_addon_dir(path: os.PathLike[str] | str) -> int | None:
    maybe_installation_dir = extract_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        product = _DELECTABLE_DIR_NAMES.get(maybe_installation_dir.name)
        if product:
            try:
                all_versions = _extract_installed_versions_from_build_info(
                    maybe_installation_dir.parent
                )
            except FileNotFoundError:
                pass
            else:
                version = all_versions.get(product['code'])
                if version:
                    return version

        try:
            return _extract_installed_version_from_config_wtf(maybe_installation_dir)
        except FileNotFoundError:
            return None


def get_addon_dir_from_installation_dir(path_like: os.PathLike[str]) -> Path:
    return Path(path_like, *_ADDON_DIR_PARTS)


def infer_flavour_from_addon_dir(path: os.PathLike[str] | str) -> Flavour | None:
    maybe_installation_dir = extract_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        try:
            return _DELECTABLE_DIR_NAMES[maybe_installation_dir.name]['flavour']
        except KeyError:
            return None



from __future__ import annotations

import asyncio
import os
import sys
import time
from pathlib import Path

import pytest

from instawow._utils.aio import run_in_thread
from instawow._utils.dist_metadata import (
    _iter_dist_infos,
    _parse_entry_points_txt,
    iter_entry_point_plugins,
)
from instawow._utils.iteration import bucketise, merge_intersecting_sets
from instawow._utils.text import tabulate
from instawow._utils.web import file_uri_to_path
from instawow._version import get_version


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


def test_tabulate_spits_out_ascii_table():
    data = [('key', 'value'), ('abc def', 'hhhdhhdhfhh'), ('dskfjsdfksdkf', 'huh')]
    assert (
        tabulate(data)
        == """\
key            value      \

-------------  -----------
abc def        hhhdhhdhfhh
dskfjsdfksdkf  huh        \
"""
    )


def test_merge_intersecting_sets_in_noncontiguous_collection():
    collection = [
        {'a'},
        {'b', 'c'},
        {'a', 'd'},
        {'e'},
        {'c', 'f'},
        {'g', 'a'},
    ]
    output = [
        {'a', 'd', 'g'},
        {'b', 'c', 'f'},
        {'e'},
    ]
    assert sorted(merge_intersecting_sets(collection)) == output


@pytest.mark.skipif(sys.platform == 'win32', reason='platform dependent')
def test_file_uri_to_path_posix_leading_slash_is_preserved():
    uri = Path('/foo/bar').as_uri()
    assert uri == 'file:///foo/bar'
    assert file_uri_to_path(uri) == '/foo/bar'


@pytest.mark.skipif(sys.platform != 'win32', reason='platform dependent')
def test_file_uri_to_path_win32_leading_slash_is_stripped():
    uri = Path('C:/foo/bar').as_uri()
    assert uri == 'file:///C:/foo/bar'
    assert file_uri_to_path(uri) == 'C:/foo/bar'


async def test_generator_in_run_in_thread_does_not_lock_up_loop():
    def foo():
        time.sleep(2)
        yield 'foo'

    async def bar():
        await asyncio.sleep(1)
        return ['bar']

    assert [
        await a
        for a in asyncio.as_completed(
            [
                run_in_thread(list)(foo()),
                bar(),
            ]
        )
    ] == [
        ['bar'],
        ['foo'],
    ]


def test_reading_entry_point_plugins_from_path(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    dist_info = tmp_path / f'instawow-{get_version()}.dist-info'
    dist_info.mkdir()
    with _iter_dist_infos([os.fspath(tmp_path)]) as dist_infos:
        assert [os.fspath(d) for d in dist_infos] == [os.fspath(dist_info)]

    entry_points_txt = """\
[console_scripts]
foo = foomod:main
# One which depends on extras:
foobar = foomod:main_bar [bar,baz]

# pytest plugins refer to a module, so there is no ':obj'
[pytest11]
nbval = nbval.plugin
"""
    assert list(_parse_entry_points_txt(entry_points_txt)) == [
        ('console_scripts', 'foo', 'foomod:main'),
        ('console_scripts', 'foobar', 'foomod:main_bar'),
        ('pytest11', 'nbval', 'nbval.plugin'),
    ]

    monkeypatch.setattr('sys.path', [tmp_path])
    (dist_info / 'entry_points.txt').write_text(entry_points_txt, encoding='utf-8')
    assert [n for n, _ in iter_entry_point_plugins('console_scripts')] == ['foo', 'foobar']
    assert [n for n, _ in iter_entry_point_plugins('pytest11')] == ['nbval']



from __future__ import annotations

import instawow.plugins

from ._cli import weakaura_updater_command_group


@instawow.plugins.hook
def instawow_add_commands():
    return (weakaura_updater_command_group,)



from __future__ import annotations

import asyncio
import concurrent.futures
import contextlib
import os
from collections.abc import Callable, Set
from functools import partial
from typing import Never

import aiohttp_client_cache
import diskcache

_http_cache_name = '_http_v1'


@contextlib.asynccontextmanager
async def make_cache(parent_dir: os.PathLike[str]):
    with concurrent.futures.ThreadPoolExecutor(1, '_http_cache') as executor:
        loop = asyncio.get_running_loop()

        def run_in_thread2[**P, T](fn: Callable[P, T]):
            async def wrapper(*args: P.args, **kwargs: P.kwargs):
                return await loop.run_in_executor(executor, partial(fn, *args, **kwargs))

            return wrapper

        # diskcache will only close the sqlite connection if it was initialised
        # in the same thread.
        cache = await run_in_thread2(diskcache.Cache)(os.path.join(parent_dir, _http_cache_name))

        def make_cache_wrapper(prefix: str):
            class Cache(aiohttp_client_cache.BaseCache):
                @run_in_thread2
                def bulk_delete(self, keys: Set[str]):
                    for key in keys:
                        cache.delete((prefix, key))

                @run_in_thread2
                def contains(self, key: str):
                    return (prefix, key) in cache

                @run_in_thread2
                def delete(self, key: str):
                    cache.delete((prefix, key))

                @run_in_thread2
                def read(self, key: str):
                    return cache.get((prefix, key))

                @run_in_thread2
                def write(self, key: str, item: aiohttp_client_cache.ResponseOrKey):
                    cache[prefix, key] = item

                async def clear(self) -> Never:
                    raise NotImplementedError

                async def keys(self):
                    if False:  # pragma: no cover
                        yield
                    raise NotImplementedError

                async def values(self):
                    if False:  # pragma: no cover
                        yield
                    raise NotImplementedError

                async def size(self) -> Never:
                    raise NotImplementedError

            return Cache()

        try:
            yield make_cache_wrapper('responses'), make_cache_wrapper('redirects')
        finally:
            await run_in_thread2(cache.close)()



"""
Utilities for reading distribution metadata from $PYTHONPATH
on the filesystem.

This lacks support for zip and other exotic paths, and `egg-info`s, which
are non-standard.

Spec: https://packaging.python.org/en/latest/specifications/entry-points/
"""

from __future__ import annotations

import importlib
import os
import sys
from contextlib import ExitStack, contextmanager
from functools import partial, reduce
from pathlib import Path


class DistNotFoundError(Exception):
    pass


@contextmanager
def _iter_dist_infos(paths: list[str]):
    with ExitStack() as exit_stack:
        yield (
            p
            for e in map(Path, paths)
            if e.is_dir()
            for p in exit_stack.enter_context(os.scandir(e or '.'))
            if p.name.endswith('.dist-info')
        )


def _parse_entry_points_txt(content: str):
    group = None
    for line in filter(None, map(str.strip, content.splitlines())):
        if line.startswith('#'):
            continue
        elif line.startswith('[') and line.endswith(']'):
            group = line.strip('[]')
        else:
            extras_pos: int = line.find('[')
            if extras_pos > -1:
                line = line[:extras_pos]
            name, _, object_ref = map(str.strip, line.partition('='))
            if object_ref:
                yield group, name, object_ref


def _import_plugin(object_ref: str):
    module, _, qualified_name = map(str.strip, object_ref.partition(':'))
    plugin = importlib.import_module(module)
    if qualified_name:
        plugin = reduce(getattr, qualified_name.split('.'), plugin)
    return plugin


def iter_entry_point_plugins(wanted_group: str):
    with _iter_dist_infos(sys.path) as dist_infos:
        for dist_info in dist_infos:
            entry_points_path = Path(dist_info.path, 'entry_points.txt')
            if entry_points_path.is_file():
                for group, name, object_ref in _parse_entry_points_txt(
                    entry_points_path.read_text(encoding='utf-8')
                ):
                    if group == wanted_group:
                        yield name, partial(_import_plugin, object_ref)


def get_version(wanted_dist_name: str):
    import email

    with _iter_dist_infos(sys.path) as dist_infos:
        for dist_info in dist_infos:
            dist_name, *_ = dist_info.name.partition('-')
            if dist_name == wanted_dist_name:
                break
        else:
            raise DistNotFoundError

    metadata = email.message_from_bytes(
        Path(dist_info.path, 'METADATA').read_bytes(),
    )
    return metadata['Version']



from __future__ import annotations

from collections.abc import Callable, Hashable, Iterable, Iterator, Set
from itertools import chain, groupby, islice, repeat
from weakref import WeakValueDictionary


class WeakValueDefaultDictionary[T, U](WeakValueDictionary[T, U]):
    def __init__(self, default_factory: Callable[[], U]) -> None:
        super().__init__()
        self.__default_factory = default_factory

    def __getitem__(self, key: T) -> U:
        try:
            return super().__getitem__(key)
        except KeyError:
            default = self[key] = self.__default_factory()
            return default


def all_eq(it: Iterable[object]) -> bool:
    "Check that all elements of an iterable are equal."
    groups = groupby(it)
    return next(groups, True) and not next(groups, False)


def bucketise[T, U](iterable: Iterable[U], key: Callable[[U], T]) -> dict[T, list[U]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket = dict[T, list[U]]()
    for value in iterable:
        bucket.setdefault(key(value), []).append(value)

    return bucket


def fill[T](it: Iterable[T], fill: T, length: int) -> Iterable[T]:
    "Fill an iterable of specified length."
    return islice(chain(it, repeat(fill)), 0, length)


def merge_intersecting_sets[T](it: Iterable[Set[T]]) -> Iterator[frozenset[T]]:
    "Recursively merge intersecting sets in a collection."
    many_sets = list(map(frozenset, it))
    while many_sets:
        this_set = many_sets.pop(0)
        while True:
            for idx, other_set in enumerate(many_sets):
                if not this_set.isdisjoint(other_set):
                    this_set |= many_sets.pop(idx)
                    break
            else:
                break
        yield this_set


def uniq[HashableT: Hashable](it: Iterable[HashableT]) -> list[HashableT]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Collection, Iterable
from functools import partial, update_wrapper


async def gather[T](it: Iterable[Awaitable[T]]) -> list[T]:
    return await asyncio.gather(*it)


def run_in_thread[**P, T](fn: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    return update_wrapper(partial(asyncio.to_thread, fn), fn)


async def cancel_tasks(tasks: Collection[asyncio.Task[object]]) -> None:
    for task in tasks:
        if not task.done():
            task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)



from __future__ import annotations

from collections.abc import Mapping
from functools import partial
from pathlib import Path
from typing import Self

from attrs import field

from instawow._utils.attrs import fauxfrozen
from instawow.config import Dirs, ProfileConfig, SecretStr, config_converter, make_plugin_dirs
from instawow.config._helpers import (
    FieldMetadata,
    ensure_dirs,
    read_config,
    read_env_vars,
    write_config,
)

from . import NAME


@fauxfrozen(kw_only=True)
class _AccessTokens:
    wago: SecretStr | None = field(
        default=None, metadata=FieldMetadata(env_prefix=True, store=True)
    )


@fauxfrozen(kw_only=True)
class PluginConfig:
    profile_config: ProfileConfig
    access_tokens: _AccessTokens = field(
        default=_AccessTokens(), metadata=FieldMetadata(env_prefix=NAME, store=True)
    )
    dirs: Dirs = field(factory=partial(make_plugin_dirs, NAME), init=False)

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, profile_config: ProfileConfig, *, env: bool = True) -> Self:
        return cls.from_values(
            read_config(cls, cls(profile_config=profile_config).config_file_path, missing_ok=True)
            | {'profile_config': profile_config},
            env=env,
        )

    def ensure_dirs(self) -> Self:
        ensure_dirs(
            self.dirs.config,
            self.profile_cache_path,
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file_path)
        return self

    @property
    def profile_cache_path(self) -> Path:
        return self.dirs.cache / 'profiles' / self.profile_config.profile

    @property
    def config_file_path(self) -> Path:
        return self.dirs.config / 'config.json'



from __future__ import annotations

import contextvars as cv
from collections.abc import Callable
from contextlib import AbstractContextManager
from itertools import chain
from typing import Self

from . import _sources, pkg_db, plugins
from . import config as _config
from . import resolvers as _resolvers
from ._utils.attrs import fauxfrozen

_config_party_var: cv.ContextVar[ConfigParty | Callable[[], _config.ProfileConfig]] = (
    cv.ContextVar('_config_party_var')
)


@fauxfrozen
class ConfigParty:
    config: _config.ProfileConfig
    database: AbstractContextManager[pkg_db.Connection]
    resolvers: _resolvers.Resolvers

    @classmethod
    def from_config(cls, config: _config.ProfileConfig) -> Self:
        return cls(config, _ReentrantDatabaseHandle(), _make_resolvers())


class _ReentrantDatabaseHandle(AbstractContextManager['pkg_db.Connection']):
    def __init__(self) -> None:
        self._connection = None
        self._referent_count = 0

    def __enter__(self) -> pkg_db.Connection:
        if self._connection is None:
            self._connection = pkg_db.prepare_database(config().db_file_path)
        self._referent_count += 1
        return self._connection

    def __exit__(self, *_: object) -> None:
        self._referent_count -= 1
        if self._referent_count == 0 and self._connection is not None:
            self._connection.close()
            self._connection = None


def _make_resolvers():
    return _resolvers.Resolvers(
        r() if callable(r) else r
        for r in chain(
            (r for g in plugins.get_plugin_resolvers() for r in g),
            _sources.DEFAULT_RESOLVERS,
        )
    )


def _get_config_party():
    config_party = _config_party_var.get()
    if callable(config_party):
        config_party = ConfigParty.from_config(config_party())
        _config_party_var.set(config_party)
    return config_party


@object.__new__
class config:
    def __call__(self) -> _config.ProfileConfig:
        return _get_config_party().config

    set = staticmethod(_config_party_var.set)
    reset = staticmethod(_config_party_var.reset)


def database() -> AbstractContextManager[pkg_db.Connection]:
    return _get_config_party().database


def resolvers() -> _resolvers.Resolvers:
    return _get_config_party().resolvers



from __future__ import annotations


def get_checksum(*values: object) -> str:
    "Base-16-encode a string using SHA-256, truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(str(v) for v in values if v).encode()).hexdigest()[:32]



from __future__ import annotations

import re
from collections.abc import Sequence
from datetime import UTC, datetime, timedelta
from itertools import takewhile
from typing import Literal, NotRequired

from typing_extensions import TypedDict
from yarl import URL

from .. import http_ctx
from .._logging import logger
from .._utils.aio import gather
from .._utils.iteration import uniq
from .._utils.text import normalise_names
from .._utils.web import as_plain_text_data_url
from ..definitions import ChangelogFormat, Defn, SourceMetadata
from ..resolvers import BaseResolver, CatalogueEntryCandidate, PkgCandidate
from ..results import PkgNonexistent, resultify
from ..wow_installations import Flavour, FlavourVersions, to_flavour

_slugify = normalise_names('-')


class _WowiCommonTerms(TypedDict):
    UID: str  # Unique add-on ID
    UICATID: str  # ID of category add-on is placed in
    UIVersion: str  # Add-on version
    UIDate: int  # Upload date expressed as unix epoch
    UIName: str  # User-facing add-on name
    UIAuthorName: str


class _WowiListApiItem(_WowiCommonTerms):
    UIFileInfoURL: str  # Add-on page on WoWI
    UIDownloadTotal: str  # Total number of downloads
    UIDownloadMonthly: str  # Number of downloads in the last month and not 'monthly'
    UIFavoriteTotal: str
    UICompatibility: list[_WowiListApiItem_CompatibilityEntry] | None  # ``null`` if would be empty
    UIDir: list[str]  # Names of folders contained in archive
    UIIMG_Thumbs: list[str] | None  # Thumbnail URLs; ``null`` if would be empty
    UIIMGs: list[str] | None  # Full-size image URLs; ``null`` if would be empty
    # There are only two add-ons on the entire list with siblings
    # (they refer to each other). I don't know if this was meant to capture
    # dependencies (probably not) but it's so underused as to be worthless.
    # ``null`` if would be empty
    UISiblings: list[str] | None
    UIDonationLink: NotRequired[str | None]  # Absent from the first item on the list (!)


class _WowiListApiItem_CompatibilityEntry(TypedDict):
    version: str  # Game version, e.g. '8.3.0'
    name: str  # Xpac or patch name, e.g. "Visions of N'Zoth" for 8.3.0


class _WowiDetailsApiItem(_WowiCommonTerms):
    UIMD5: str | None  # Archive hash, ``null` when review pending
    UIFileName: str  # The actual filename, e.g. 'foo.zip'
    UIDownload: str  # Download URL
    UIPending: Literal['0', '1']  # Set to '1' if the file is awaiting approval
    UIDescription: str  # Long description with BB Code and all
    UIChangeLog: str  # This can also contain BB Code
    UIHitCount: str  # Same as UIDownloadTotal
    UIHitCountMonthly: str  # Same as UIDownloadMonthly


def _timestamp_to_datetime(timestamp: int):
    return datetime.fromtimestamp(timestamp / 1000, UTC)


class WowiResolver(BaseResolver[_WowiDetailsApiItem]):
    metadata = SourceMetadata(
        id='wowi',
        name='WoWInterface',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Raw,
        addon_toc_key='X-WoWI-ID',
    )

    # Ref: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the unreleased Minion v4,
    # which is fair to assume is unstable.  The naming scheme was changed to
    # camel case and some fields which contained strings were turned into numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # ``UICompatibility`` can't be relied on to enforce compatibility - not only
    # because it's not consistently populated, but also because it's not
    # reported by the details endpoint and keeping the two endpoints
    # in sync is tricky.
    # We could theoretically inspect the downloaded file for compatbility,
    # but that's a lot of bandwidth to be wasting on add-ons which might
    # not end up being installed.
    # The API appears to inherit the version of the latest
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever, the download link always points to the
    # 'retail' version.
    # In essence, the WoWI API is only good for installing retail add-ons.
    __list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    __details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    def get_alias_from_url(self, url: URL):
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                return url.query.get('fileid')
            elif url.name == 'fileinfo.php':
                return url.query.get('id')
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                return match and match['id']

    async def resolve(self, defns: Sequence[Defn]):
        defn_ids = {d: ''.join(takewhile(str.isdigit, d.alias)) for d in defns}

        async with http_ctx.web_client().get(
            (self.__details_api_url / f'{",".join(uniq(filter(None, defn_ids.values())))}.json'),
            expire_after=timedelta(minutes=15),
        ) as response:
            if response.status == 404:
                return await super().resolve(defns)

            response.raise_for_status()

            addons_details = {i['UID']: i for i in await response.json()}

        resolve_one = resultify(self.resolve_one)
        results = await gather(resolve_one(d, addons_details.get(i)) for d, i in defn_ids.items())
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _WowiDetailsApiItem | None):
        if metadata is None:
            raise PkgNonexistent

        return PkgCandidate(
            id=metadata['UID'],
            slug=_slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=f'https://www.wowinterface.com/downloads/info{metadata["UID"]}',
            download_url=metadata['UIDownload'],
            date_published=_timestamp_to_datetime(metadata['UIDate']),
            version=metadata['UIVersion'],
            changelog_url=as_plain_text_data_url(metadata['UIChangeLog']),
        )

    async def catalogue(self):
        logger.debug(f'Retrieving {self.__list_api_url}')

        async with http_ctx.web_client().get(
            self.__list_api_url, raise_for_status=True
        ) as response:
            items: list[_WowiListApiItem] = await response.json()

        supported_flavours = frozenset(Flavour)

        for item in items:
            match item:
                case {'UICompatibility': list(compatibility)}:
                    yield CatalogueEntryCandidate(
                        id=item['UID'],
                        name=item['UIName'],
                        url=item['UIFileInfoURL'],
                        game_flavours=supported_flavours.intersection(
                            to_flavour(f)
                            for c in compatibility
                            for f in (FlavourVersions.from_version_string(c['version']),)
                            if f
                        ),
                        download_count=int(item['UIDownloadTotal']),
                        last_updated=_timestamp_to_datetime(item['UIDate']),
                        folders=[frozenset(item['UIDir'])],
                    )

                case _:
                    continue



from __future__ import annotations

from collections.abc import Iterable
from functools import lru_cache
from typing import Protocol

import click
import pluggy

from . import NAME, resolvers
from ._utils.dist_metadata import iter_entry_point_plugins

_hookspec = pluggy.HookspecMarker(NAME)
_hookimpl = pluggy.HookimplMarker(NAME)

# Keep "hookimpl" for back-compat.
hook = hookimpl = _hookimpl


class InstawowPlugin(Protocol):  # pragma: no cover
    "The plug-in interface."

    @_hookspec
    def instawow_add_commands(self) -> Iterable[click.Command]:
        "Additional commands to register with ``click``."
        ...

    @_hookspec
    def instawow_add_resolvers(self) -> Iterable[resolvers.Resolver]:
        "Additional resolvers to load."
        ...


@lru_cache(1)
def _load_plugins():
    plugin_manager = pluggy.PluginManager(NAME)
    plugin_manager.add_hookspecs(InstawowPlugin)

    for name, load_plugin in iter_entry_point_plugins(f'{NAME}.plugins'):
        plugin_manager.register(load_plugin(), name=name)

    return plugin_manager.hook


def get_plugin_commands() -> Iterable[Iterable[click.Command]]:
    return _load_plugins().instawow_add_commands()


def get_plugin_resolvers() -> Iterable[Iterable[resolvers.Resolver]]:
    return _load_plugins().instawow_add_resolvers()



from __future__ import annotations

import os
import sys
from functools import partial
from pathlib import Path

from .. import NAME

_get_home_dir = partial(os.environ.get, f'{NAME.upper()}_HOME')


def get_config_dir(*parts: str):
    home = _get_home_dir()
    if home:
        return Path(home, 'config', *parts)

    parent_dir = os.environ.get('XDG_CONFIG_HOME')

    if not parent_dir:
        if sys.platform == 'darwin':
            parent_dir = Path.home() / 'Library' / 'Application Support'
        elif sys.platform == 'win32':
            parent_dir = os.environ.get('APPDATA')

    if not parent_dir:
        parent_dir = Path.home() / '.config'

    return Path(parent_dir, NAME, *parts)


def get_cache_dir(*parts: str):
    home = _get_home_dir()
    if home:
        return Path(home, 'cache', *parts)

    parent_dir = os.environ.get('XDG_CACHE_HOME')

    if not parent_dir:
        if sys.platform == 'darwin':
            parent_dir = Path.home() / 'Library' / 'Caches'
        elif sys.platform == 'win32':
            parent_dir = os.environ.get('LOCALAPPDATA')

    if not parent_dir:
        parent_dir = Path.home() / '.cache'

    return Path(parent_dir, NAME, *parts)


def get_state_dir(*parts: str):
    home = _get_home_dir()
    if home:
        return Path(home, 'state', *parts)

    parent_dir = os.environ.get('XDG_STATE_HOME')

    if not parent_dir and sys.platform not in {'darwin', 'win32'}:
        parent_dir = Path.home() / '.local' / 'state'

    if not parent_dir:
        return get_config_dir()

    return Path(parent_dir, NAME, *parts)



from __future__ import annotations

import importlib.resources
import json
import re
from functools import cache
from io import BytesIO
from zipfile import ZipFile

from instawow._version import get_version

from ._mock_server import AddRoutes as AddRoutes
from ._mock_server import Response
from ._mock_server import Route as Route
from ._mock_server import patch_aiohttp as patch_aiohttp
from ._mock_server import prepare_mock_server_router as prepare_mock_server_router


def _load_fixture(filename: str):
    return (importlib.resources.files(__spec__.parent) / filename).read_bytes()


def _load_json_fixture(filename: str):
    return json.loads(_load_fixture(filename))


@cache
def _make_addon_zip(*folders: str):
    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        for folder in folders:
            file.writestr(f'{folder}/{folder}.toc', b'')

    return buffer.getvalue()


ROUTES = {
    r.url: r
    for r in (
        Route(
            r'//pypi\.org/pypi/instawow/json',
            {'info': {'version': get_version()}},
        ),
        Route(
            r'//raw\.githubusercontent\.com/layday/instawow-data/data/base-catalogue-v8\.compact\.json',
            _load_json_fixture('base-catalogue-v8.compact.json'),
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods',
            _load_json_fixture('curse-addon--all.json'),
            method='POST',
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods/search\?gameId=1&slug=masque',
            _load_json_fixture('curse-addon-slug-search.json'),
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods/13592/files',
            _load_json_fixture('curse-addon-files.json'),
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods/13592/files/6454541',
            _load_json_fixture('curse-addon-file-6454541.json'),
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods/13592/files/5810397',
            _load_json_fixture('curse-addon-file-5810397.json'),
        ),
        Route(
            r'//api\.curseforge\.com/v1/mods/13592/files/(\d+)/changelog',
            _load_json_fixture('curse-addon-changelog.json'),
        ),
        Route(
            r'//edge\.forgecdn\.net/.*',
            lambda: Response(body=_make_addon_zip('Masque')),
        ),
        Route(
            r'//api\.mmoui\.com/v3/game/WOW/filelist\.json',
            _load_json_fixture('wowi-filelist.json'),
        ),
        Route(
            r'//api\.mmoui\.com/v3/game/WOW/filedetails/(\d*)\.json',
            _load_json_fixture('wowi-filedetails.json'),
        ),
        Route(
            r'//cdn\.wowinterface\.com/.*',
            lambda: Response(body=_make_addon_zip('Masque')),
        ),
        Route(
            r'//api\.tukui\.org/v1/addon/tukui',
            _load_json_fixture('tukui-ui--tukui.json'),
        ),
        Route(
            r'//api\.tukui\.org/v1/addon/elvui',
            _load_json_fixture('tukui-ui--elvui.json'),
        ),
        Route(
            r'//api\.tukui\.org/v1/download/.*',
            lambda: Response(body=_make_addon_zip('Tukui')),
        ),
        Route(
            r'//api\.github\.com/repos/nebularg/PackagerTest',
            _load_json_fixture('github-repo-release-json.json'),
        ),
        Route(
            r'//api\.github\.com/repos/nebularg/PackagerTest/releases\?per_page=10',
            _load_json_fixture('github-release-release-json.json'),
        ),
        Route(
            r'//api\.github\.com/repos/nebularg/PackagerTest/releases/assets/37156458',
            _load_json_fixture('github-release-release-json-release-json.json'),
        ),
        Route(
            r'//api\.github\.com/repositories/44074003',
            _load_json_fixture('github-repo-masque.json'),
        ),
        Route(
            r'//api\.github\.com/repos/SFX-WoW/Masque',
            _load_json_fixture('github-repo-masque.json'),
        ),
        Route(
            r'//api\.github\.com/repositories/44074003/releases\?per_page=10',
            _load_json_fixture('github-release-masque.json'),
        ),
        Route(
            r'//api\.github\.com/repos/SFX-WoW/Masque/releases\?per_page=10',
            _load_json_fixture('github-release-masque.json'),
        ),
        Route(
            re.escape(
                next(
                    a['url']
                    for r in _load_json_fixture('github-release-masque.json')
                    if not r['prerelease']
                    for a in r['assets']
                    if a['name'] == 'release.json'
                )
            ),
            _load_json_fixture('github-release-masque-release-json.json'),
        ),
        Route(
            r'//api\.github\.com/repos/AdiAddons/AdiBags',
            _load_json_fixture('github-repo-no-releases.json'),
        ),
        Route(
            r'//api\.github\.com/repos/AdiAddons/AdiBags/releases\?per_page=10',
            lambda: Response(body=b'', status=404),
        ),
        Route(
            r'//api\.github\.com/repos/AdiAddons/AdiButtonAuras/releases/tags/2\.0\.19',
            _load_json_fixture('github-release-no-assets.json'),
        ),
        Route(
            r'//api\.github\.com/repos/layday/foobar',
            lambda: Response(body=b'', status=404),
        ),
        Route(
            r'//api\.github\.com/repos(/[^/]*){2}/releases/assets/.*',
            lambda: Response(body=_make_addon_zip('Masque')),
        ),
        Route(
            r'//github\.com/login/device/code',
            _load_json_fixture('github-oauth-login-device-code.json'),
            method='POST',
        ),
        Route(
            r'//github\.com/login/oauth/access_token',
            _load_json_fixture('github-oauth-login-access-token.json'),
            method='POST',
        ),
        Route(
            r'//api\.github\.com/repos/28/NoteworthyII',
            _load_json_fixture('github-repo-no-release-json.json'),
        ),
        Route(
            r'//api\.github\.com/repos/28/NoteworthyII/releases\?per_page=10',
            _load_json_fixture('github-release-no-release-json.json'),
        ),
    )
}



from __future__ import annotations

from typing import TypedDict

from . import NAME, http_ctx


def get_version() -> str:
    from ._utils.dist_metadata import DistNotFoundError, get_version

    try:
        return get_version(NAME)
    except DistNotFoundError:
        return '0+dev'


async def is_outdated(current_version: str | None = None) -> tuple[bool, str]:
    """Check on PyPI to see if instawow is outdated.

    The response is cached for 24 hours.
    """
    from datetime import timedelta

    from aiohttp import ClientError
    from packaging.version import Version

    __version__ = current_version or get_version()
    parsed_version = Version(__version__)
    if parsed_version.local:
        return (False, '')

    try:
        async with http_ctx.web_client().get(
            'https://pypi.org/simple/instawow',
            expire_after=timedelta(days=1),
            headers={
                'Accept': 'application/vnd.pypi.simple.v1+json',
            },
            raise_for_status=True,
        ) as response:
            metadata: TypedDict[{'versions': list[str]}] = await response.json()

    except ClientError:
        version = __version__
    else:
        version = max(metadata['versions'], key=Version)

    return (Version(version) > parsed_version, version)



from __future__ import annotations

import os
from collections.abc import Iterator, Mapping, Sized
from functools import partial
from pathlib import Path
from typing import Self

import attrs
from attrs import field

from .. import NAME
from .._utils.attrs import enrich_validator_exc, fauxfrozen
from .._utils.file import expand_path, trash
from ..wow_installations import Track, extract_installation_dir_from_addon_dir
from ._dirs import get_cache_dir, get_config_dir, get_state_dir
from ._helpers import (
    FieldMetadata,
    SecretStr,
    ensure_dirs,
    read_config,
    read_env_vars,
    write_config,
)
from ._helpers import UninitialisedConfigError as UninitialisedConfigError
from ._helpers import config_converter as config_converter

_path_field = partial(field, converter=expand_path)


@enrich_validator_exc
def _validate_path_is_writable_dir(_model: object, _attr: attrs.Attribute[Path], value: Path):
    if not (value.is_dir() and os.access(value, os.W_OK)):
        raise ValueError(f'"{value}" is not a writable directory')


def _make_validate_min_length(min_length: int):
    @enrich_validator_exc
    def _validate_min_length[SizedT: Sized](
        _model: object, _attr: attrs.Attribute[SizedT], value: SizedT
    ):
        if len(value) < min_length:
            raise ValueError(f'Value must have a minimum length of {min_length}')

    return _validate_min_length


@fauxfrozen(kw_only=True)
class _AccessTokens:
    cfcore: SecretStr | None = field(
        default=None,
        metadata=FieldMetadata(env_prefix=True, store=True),
    )
    github: SecretStr | None = field(
        default=None,
        metadata=FieldMetadata(env_prefix=True, store=True),
    )
    wago_addons: SecretStr | None = field(
        default=None,
        metadata=FieldMetadata(env_prefix=True, store=True),
    )


@fauxfrozen(kw_only=True)
class Dirs:
    cache: Path = _path_field()
    config: Path = _path_field()
    state: Path = _path_field()


def _make_default_dirs() -> Dirs:
    return Dirs(cache=get_cache_dir(), config=get_config_dir(), state=get_state_dir())


def make_plugin_dirs(plugin_name: str) -> Dirs:
    parts = ('plugins', plugin_name)
    return Dirs(
        cache=get_cache_dir(*parts), config=get_config_dir(*parts), state=get_state_dir(*parts)
    )


@fauxfrozen(kw_only=True)
class GlobalConfig:
    auto_update_check: bool = field(
        default=True, metadata=FieldMetadata(env_prefix=NAME, store=True)
    )
    access_tokens: _AccessTokens = field(
        default=_AccessTokens(), metadata=FieldMetadata(env_prefix=NAME, store='independently')
    )
    dirs: Dirs = field(factory=_make_default_dirs, init=False)

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, *, env: bool = True) -> Self:
        env_config = cls.from_values(env=env)
        disk_config_values = read_config(cls, env_config.config_file_path, missing_ok=True)
        return cls.from_values(disk_config_values, env=env) if disk_config_values else env_config

    def ensure_dirs(self) -> Self:
        ensure_dirs(*attrs.astuple(self.dirs))
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file_path)
        return self

    @property
    def config_file_path(self) -> Path:
        return self.dirs.config / 'config.json'


@fauxfrozen(kw_only=True)
class ProfileConfig:
    global_config: GlobalConfig
    profile: str = field(
        converter=str.strip,
        validator=_make_validate_min_length(1),
        metadata=FieldMetadata(store=True),
    )
    addon_dir: Path = _path_field(
        validator=_validate_path_is_writable_dir, metadata=FieldMetadata(store=True)
    )
    track: Track = field(
        metadata=FieldMetadata(
            store=True,
            store_alias='game_flavour',  # For backward compatibility.
        )
    )

    @classmethod
    def _make_config_file_path(cls, global_config: GlobalConfig, profile: str):
        return global_config.dirs.config / 'profiles' / profile / 'config.json'

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, env: bool = False) -> Self:
        if env:
            values = read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, global_config: GlobalConfig, profile: str, *, env: bool = True) -> Self:
        return cls.from_values(
            read_config(cls, cls._make_config_file_path(global_config, profile))
            | {'global_config': global_config},
            env=env,
        )

    @classmethod
    def iter_profiles(cls, global_config: GlobalConfig) -> Iterator[str]:
        yield from (
            c.parent.name for c in global_config.dirs.config.glob('profiles/*/config.json')
        )

    @classmethod
    def iter_profile_installations(cls, global_config: GlobalConfig) -> Iterator[Path]:
        for config_json in global_config.dirs.config.glob('profiles/*/config.json'):
            addon_dir = read_config(cls, config_json)['addon_dir']
            maybe_installation_dir = extract_installation_dir_from_addon_dir(addon_dir)
            if maybe_installation_dir:
                yield maybe_installation_dir

    def ensure_dirs(self) -> Self:
        ensure_dirs(
            self.config_path,
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        write_config(self, self.config_file_path)
        return self

    def delete(self) -> None:
        trash((self.config_path,))

    @property
    def config_path(self) -> Path:
        return self.config_file_path.parent

    @property
    def config_file_path(self) -> Path:
        return self._make_config_file_path(self.global_config, self.profile)

    @property
    def db_file_path(self) -> Path:
        return self.config_path / 'db.sqlite'



from __future__ import annotations

from collections.abc import Set
from datetime import datetime
from functools import cached_property
from typing import Any, Self

import cattrs
import cattrs.preconf.json

from .. import config_ctx
from .._utils.attrs import fauxfrozen
from .._utils.iteration import bucketise
from .._utils.text import normalise_names
from ..wow_installations import Flavour

CATALOGUE_VERSION = 8


_catalogue_converter = cattrs.Converter(
    unstruct_collection_overrides={
        Set: sorted,
    }
)
cattrs.preconf.json.configure_converter(_catalogue_converter)

_normalise_name = normalise_names('')


@fauxfrozen(kw_only=True)
class AddonKey:
    source: str
    id: str


@fauxfrozen(kw_only=True)
class CatalogueEntry(AddonKey):
    slug: str
    name: str
    url: str
    game_flavours: frozenset[Flavour]
    download_count: int
    last_updated: datetime
    folders: list[frozenset[str]]
    same_as: list[AddonKey]
    normalised_name: str
    derived_download_score: float


async def collate(start_date: datetime | None) -> dict[str, Any]:
    return _catalogue_converter.unstructure(
        {
            'version': CATALOGUE_VERSION,
            'entries': [
                {
                    'source': r.metadata.id,
                    'slug': '',
                    'folders': [],
                    'same_as': [],
                }
                | e
                for r in config_ctx.resolvers().values()
                async for e in r.catalogue()
                if not start_date or e['last_updated'] >= start_date
            ],
        }
    )


@fauxfrozen(kw_only=True)
class ComputedCatalogue:
    entries: list[CatalogueEntry]

    @classmethod
    def from_base_catalogue(cls, unstructured_base_catalogue: dict[str, Any]) -> Self:
        from .._sources.github import GithubResolver

        normalise_name = _normalise_name

        base_entries = unstructured_base_catalogue['entries']

        most_downloads_per_source = {
            s: max(e['download_count'] for e in i) or 1
            for s, i in bucketise(base_entries, key=lambda e: e['source']).items()
        }
        same_as_from_github = {
            (s['source'], s['id']): [
                e,
                *(i for i in e['same_as'] if i['source'] != s['source']),
            ]
            for e in base_entries
            if e['source'] == GithubResolver.metadata.id and e['same_as']
            for s in e['same_as']
        }

        return _catalogue_converter.structure(
            {
                'entries': (
                    e
                    | {
                        'same_as': e['same_as']
                        if e['source'] == GithubResolver.metadata.id
                        else (same_as_from_github.get((e['source'], e['id'])) or e['same_as']),
                        'normalised_name': normalise_name(e['name']),
                        'derived_download_score': e['download_count']
                        / most_downloads_per_source[e['source']],
                    }
                    for e in base_entries
                ),
            },
            cls,
        )

    @cached_property
    def keyed_entries(self) -> dict[tuple[str, str], CatalogueEntry]:
        return {(e.source, e.id): e for e in self.entries}



from __future__ import annotations

from datetime import UTC, datetime
from typing import TypedDict

from .. import config_ctx, sync_ctx
from .._utils.aio import run_in_thread
from ..definitions import ChangelogFormat, Defn, SourceMetadata
from ..resolvers import BaseResolver, CatalogueEntryCandidate, PkgCandidate
from ..results import PkgNonexistent
from ..wow_installations import Flavour

_READ_PLUGIN_CONFIG_LOCK = (object(), '_READ_PLUGIN_CONFIG_')


class _ResolveMetadata(TypedDict):
    requires_build: bool


_ADDONS = {
    ('0', 'weakauras-companion'): _ResolveMetadata(requires_build=False),
    ('1', 'weakauras-companion-autoupdate'): _ResolveMetadata(requires_build=True),
}


class InstawowResolver(BaseResolver):
    metadata = SourceMetadata(
        id='instawow',
        name='instawow',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key=None,
    )

    __plugin_config = None

    async def resolve_one(self, defn: Defn, metadata: None):
        from instawow_weakaura_updater import builder
        from instawow_weakaura_updater.config import PluginConfig

        try:
            (id_, slug), metadata_ = next((p, v) for p, v in _ADDONS.items() if defn.alias in p)
        except StopIteration:
            raise PkgNonexistent from None

        async with sync_ctx.locks()[_READ_PLUGIN_CONFIG_LOCK]:
            plugin_config = self.__plugin_config
            if plugin_config is None:
                plugin_config = self.__plugin_config = await run_in_thread(PluginConfig.read)(
                    config_ctx.config()
                )

        build_paths = (
            await builder.build_addon(plugin_config)
            if metadata_['requires_build']
            else builder.get_build_paths(plugin_config)
        )

        return PkgCandidate(
            id=id_,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=build_paths.archive.as_uri(),
            date_published=datetime.now(UTC),
            version=await run_in_thread(build_paths.version.read_text)(encoding='utf-8'),
            changelog_url=build_paths.changelog.as_uri(),
        )

    async def catalogue(self):
        yield CatalogueEntryCandidate(
            id='1',
            slug='weakauras-companion-autoupdate',
            name='WeakAuras Companion',
            url='https://github.com/layday/instawow',
            game_flavours=frozenset(Flavour),
            download_count=1,
            last_updated=datetime.now(UTC),
            folders=[
                frozenset({'WeakAurasCompanion'}),
            ],
        )



from __future__ import annotations

import json
import shutil
from functools import partial
from textwrap import dedent
from unittest import mock

import click.testing
import prompt_toolkit.input
import pytest

from instawow.cli import cli
from instawow.config import ProfileConfig
from instawow.config._helpers import make_display_converter

pytestmark = pytest.mark.usefixtures(
    '_iw_config_ctx', '_iw_web_client_ctx', '_iw_mock_pt_progress_bar', '_iw_mock_asyncio_run'
)


_runner = click.testing.CliRunner()
run = partial(_runner.invoke, cli, catch_exceptions=False)


def install_masque():
    run('install curse:masque')


def pretend_install_masque(
    iw_profile_config: ProfileConfig,
):
    masque = iw_profile_config.addon_dir / 'Masque'
    masque.mkdir()
    (masque / 'Masque.toc').write_text(
        """\
## X-Curse-Project-ID: 13592
## X-WoWI-ID: 12097
""",
        encoding='utf-8',
    )


@pytest.mark.parametrize(
    'alias',
    [
        'curse:masque',
        'wowi:12097-masque',
        'tukui:tukui',
        'github:sfx-wow/masque',
    ],
)
def test_valid_pkg_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').stdout.startswith(f'✓ {alias}\n  installed')
    assert run(f'install {alias}').stdout == f'✗ {alias}\n  package already installed\n'
    assert run(f'update {alias}').stdout == f'✗ {alias}\n  package is up to date\n'
    assert run(f'remove {alias}').stdout == f'✓ {alias}\n  removed\n'
    assert run(f'update {alias}').stdout == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').stdout == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['instawow:gargantuan-wigs'])
def test_nonexistent_pkg_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').stdout == f'✗ {alias}\n  package does not exist\n'
    assert run(f'update {alias}').stdout == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').stdout == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('iw_global_config_values', [None], indirect=True)
@pytest.mark.parametrize('alias', ['curse:masque'])
def test_disabled_source_lifecycle(
    alias: str,
):
    assert (
        run(f'install {alias}').stdout
        == f'✗ {alias}\n  package source is disabled: access token missing\n'
    )
    assert run(f'update {alias}').stdout == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').stdout == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['foo:bar'])
def test_invalid_source_lifecycle(
    alias: str,
):
    assert run(f'install {alias}').stdout == f'✗ :{alias}\n  package source is invalid\n'
    assert run(f'update {alias}').stdout == f'✗ :{alias}\n  package is not installed\n'
    assert run(f'remove {alias}').stdout == f'✗ {alias}\n  package is not installed\n'


def test_reconciled_folder_conflict_on_install():
    assert run('install curse:masque').stdout.startswith('✓ curse:masque\n  installed')
    assert run('install wowi:12097-masque').stdout == (
        '✗ wowi:12097-masque\n'
        '  package folders conflict with installed package Masque (curse:13592)\n'
    )


def test_unreconciled_folder_conflict_on_install(
    iw_profile_config: ProfileConfig,
):
    iw_profile_config.addon_dir.joinpath('Masque').mkdir()
    assert (
        run('install curse:masque').stdout
        == "✗ curse:masque\n  package folders conflict with 'Masque'\n"
    )
    assert run('install --replace curse:masque').stdout.startswith('✓ curse:masque\n  installed')


def test_keep_folders_on_remove(
    iw_profile_config: ProfileConfig,
):
    install_masque()
    assert run('remove --keep-folders curse:masque').stdout == '✓ curse:masque\n  removed\n'
    assert iw_profile_config.addon_dir.joinpath('Masque').is_dir()


def test_version_strategy_lifecycle():
    assert run('install curse:masque').stdout.startswith(
        '✓ curse:masque\n  installed 11.1.5_6454541'
    )
    assert (
        run('install curse:masque#version_eq=foo').stdout
        == '✗ curse:masque\n  package already installed\n'
    )
    assert run('update curse:masque').stdout == '✗ curse:masque\n  package is up to date\n'
    assert run('update curse:masque#version_eq=11.0.2').stdout == dedent(
        """\
        ✓ curse:masque
          updated 11.1.5_6454541 to 11.0.2_5810397 with new strategies:
            version_eq='11.0.2_5810397'
        """
    )
    assert run('remove curse:masque').stdout == '✓ curse:masque\n  removed\n'
    assert run('install curse:masque#version_eq=foo').stdout == dedent(
        """\
        ✗ curse:masque
          no files are available for download
        """
    )
    assert (
        run('install curse:masque#version_eq=11.0.2').stdout
        == '✓ curse:masque\n  installed 11.0.2_5810397\n'
    )
    assert run('update').stdout == '✗ curse:masque\n  package is pinned\n'
    assert run('update curse:masque#=').stdout == dedent(
        """\
        ✓ curse:masque
          updated 11.0.2_5810397 to 11.1.5_6454541 with new strategies:
            version_eq=None
        """
    )
    assert run('remove curse:masque').stdout == '✓ curse:masque\n  removed\n'


def test_install_options():
    assert run('install curse:masque#any_release_type,any_flavour').stdout == dedent(
        """\
        ✓ curse:masque
          installed 11.1.7-Beta-2_6725716
        """
    )


@pytest.mark.parametrize('step', [1, -1])
def test_install_order_is_respected(
    step: int,
):
    assert run(
        'install '
        + ' '.join(
            [
                'curse:masque',
                'curse:masque#version_eq=11.0.2',
            ][::step]
        )
    ).stdout == dedent(
        f"""\
        ✓ curse:masque
          installed {'11.1.5_6454541' if step == 1 else '11.0.2_5810397'}
        ✗ curse:masque
          package folders conflict with installed package Masque (curse:13592)
        """
    )


def test_install_invalid_defn():
    result = run('install foo')
    assert result.exit_code == 2
    assert result.stderr.endswith("Error: Invalid value for '[ADDONS]...': foo\n")


def test_install_dry_run():
    assert run('install --dry-run curse:masque').stdout == dedent(
        """\
        ✓ curse:masque
          would have installed 11.1.5_6454541
        """
    )


def test_debug_config(
    iw_profile_config: ProfileConfig,
):
    assert (
        run('debug config').stdout
        == json.dumps(make_display_converter().unstructure(iw_profile_config), indent=2) + '\n'
    )


def test_debug_profile(
    iw_profile_config: ProfileConfig,
):
    assert (
        run('debug profiles').stdout
        == json.dumps(
            make_display_converter().unstructure(
                iw_profile_config.iter_profiles(iw_profile_config.global_config), list[str]
            ),
            indent=2,
        )
        + '\n'
    )


def test_debug_sources():
    import cattrs.preconf.json

    from instawow.definitions import SourceMetadata

    json_converter = cattrs.preconf.json.make_converter()

    output = run('debug sources').stdout
    source_metadata = json_converter.loads(output, list[SourceMetadata])
    assert len(source_metadata) > 1


@pytest.mark.parametrize('command', ['configure', 'list'], ids=['explicit', 'implicit'])
def test_configure__create_new_profile(
    monkeypatch: pytest.MonkeyPatch,
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
    command: str,
):
    # In case there is a WoW installation in the test environment.
    monkeypatch.setattr('instawow.wow_installations.find_installations', lambda: iter([]))

    iw_pt_input.send_text(f'{iw_profile_config.addon_dir}\r\rY\r')
    assert run(f'-p foo {command}').stdout == (
        'Generate an access token for GitHub to avoid being rate limited. You\n'
        'are only allowed to perform 60 requests an hour without one.\n'
        'Navigate to https://github.com/login/device and paste the code below:\n'
        '  WDJB-MJHT\n'
        'Waiting...\n'
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.dirs.config / "config.json"}\n'
        f'  {iw_profile_config.global_config.dirs.config / "profiles/foo/config.json"}\n'
    )


def test_configure__update_existing_profile_interactively(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    iw_pt_input.send_text('Y\r')
    assert run('configure global_config.auto_update_check').stdout == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.dirs.config / "config.json"}\n'
        f'  {iw_profile_config.global_config.dirs.config / "profiles/__default__/config.json"}\n'
    )


def test_configure__update_existing_profile_directly(
    iw_profile_config: ProfileConfig,
):
    assert run('configure global_config.auto_update_check=0').stdout == (
        'Configuration written to:\n'
        f'  {iw_profile_config.global_config.dirs.config / "config.json"}\n'
        f'  {iw_profile_config.global_config.dirs.config / "profiles/__default__/config.json"}\n'
    )


def test_erase_profile(
    iw_profile_config: ProfileConfig,
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    assert iw_profile_config.config_path.exists()
    iw_pt_input.send_text('Y\r')
    assert run('profile erase').stdout == 'Profile erased.\n'
    assert not iw_profile_config.config_path.exists()


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__pkg_not_installed(
    options: str,
):
    assert (
        run(f'rollback {options} curse:masque').stdout
        == '✗ curse:masque\n  package is not installed\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__unsupported(
    options: str,
):
    assert run('install wowi:12097-masque').exit_code == 0
    assert (
        run(f'rollback {options} wowi:12097-masque').stdout
        == '✗ wowi:12097-masque\n  strategies are not valid for source: version_eq\n'
    )


def test_rollback__single_version():
    assert run('install curse:masque').exit_code == 0
    assert run('rollback curse:masque').stdout == '✗ curse:masque\n  cannot find older versions\n'


def test_rollback__multiple_versions(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    assert run('install curse:masque#version_eq=11.0.2').exit_code == 0
    assert run('remove curse:masque').exit_code == 0
    assert run('install curse:masque').exit_code == 0
    iw_pt_input.send_text('\r\r')
    assert run('rollback curse:masque').stdout == dedent(
        """\
        ✓ curse:masque
          updated 11.1.5_6454541 to 11.0.2_5810397 with new strategies:
            version_eq='11.0.2_5810397'
        """
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__rollback_multiple_versions(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    options: str,
):
    assert run('install curse:masque').exit_code == 0
    assert run('remove curse:masque').exit_code == 0
    assert run('install curse:masque#version_eq=11.0.2').exit_code == 0
    iw_pt_input.send_text('\r\r')
    assert run(f'rollback {options} curse:masque').stdout == dedent(
        """\
        ✓ curse:masque
          updated 11.0.2_5810397 to 11.1.5_6454541 with new strategies:
            version_eq=None
        """
        if options == '--undo'
        else """\
        ✓ curse:masque
          updated 11.0.2_5810397 to 11.1.5_6454541
        """
    )


def test_reconcile__list_unreconciled(
    iw_profile_config: ProfileConfig,
):
    pretend_install_masque(iw_profile_config)
    assert run('reconcile --list-unreconciled').stdout == dedent(
        """\
        unreconciled
        ------------
        Masque      \

        """
    )


def test_reconcile_leftovers(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_masque(iw_profile_config)
    iw_pt_input.send_text('sss')  # Skip
    assert run('reconcile').stdout.endswith(
        dedent(
            """\
            unreconciled
            ------------
            Masque      \

            """
        )
    )


def test_reconcile__auto_reconcile(
    iw_profile_config: ProfileConfig,
):
    pretend_install_masque(iw_profile_config)
    assert run('reconcile --auto').stdout == dedent(
        """\
        ✓ github:sfx-wow/masque
          installed 11.1.5
        """
    )


def test_reconcile__abort_interactive_reconciliation(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_masque(iw_profile_config)
    iw_pt_input.send_text('\x03')  # ^C
    assert run('reconcile').stderr.endswith('Aborted!\n')


def test_reconcile__complete_interactive_reconciliation(
    iw_pt_input: prompt_toolkit.input.PipeInput,
    iw_profile_config: ProfileConfig,
):
    pretend_install_masque(iw_profile_config)
    iw_pt_input.send_text('\ry')
    assert run('reconcile').stdout.endswith(
        dedent(
            """\
            ✓ github:sfx-wow/masque
              installed 11.1.5
            """
        )
    )


def test_reconcile__reconciliation_complete():
    assert run('reconcile').stdout == 'No add-ons left to reconcile.\n'


def test_rereconcile(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    install_masque()
    iw_pt_input.send_text('\ry')
    assert run('rereconcile').stdout == dedent(
        """\
        ✓ curse:masque
          removed
        ✓ github:sfx-wow/masque
          installed 11.1.5
        """
    )


def test_rereconcile_with_args(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    install_masque()

    iw_pt_input.send_text('\ry')
    assert run('rereconcile foo').stdout == ''

    iw_pt_input.send_text('\ry')
    assert run('rereconcile masque').stdout == dedent(
        """\
        ✓ curse:masque
          removed
        ✓ github:sfx-wow/masque
          installed 11.1.5
        """
    )


def test_search__no_results():
    assert run('search ∅').stdout == 'No results found.\n'


def test_search__exit_without_selecting(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text('\r')  # enter
    assert run('search masque').stdout == (
        'Nothing was selected; select add-ons with <space> and confirm by pressing <enter>.\n'
    )


def test_search__exit_after_selection(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \rn')  # space, enter, "n"
    assert run('search masque').stdout == ''


def test_search__install_one(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \ry')  # space, enter, enter
    assert run('search masque --source curse').stdout == dedent(
        """\
        ✓ curse:masque
          installed 11.1.5_6454541
        """
    )


def test_search__install_multiple_conflicting(
    iw_pt_input: prompt_toolkit.input.PipeInput,
):
    iw_pt_input.send_text(' \x1b[B \ry')  # space, arrow down, space, enter, enter
    assert run('search masque').stdout == dedent(
        """\
        ✓ wowi:12097
          installed 11.1.7-Beta-2
        ✗ curse:masque
          package folders conflict with installed package Masque (wowi:12097)
        """
    )


def test_changelog_output_installed_convert():
    install_masque()
    stdout = (
        'curse:masque:\n  11.1.5' if shutil.which('pandoc') else 'curse:masque:\n  <h2>11.1.5</h2>'
    )
    assert run('view-changelog curse:masque').stdout.startswith(stdout)


def test_changelog_output_installed_no_convert():
    install_masque()
    assert run('view-changelog --no-convert curse:masque').stdout.startswith(
        'curse:masque:\n  <h2>11.1.5</h2>'
    )


def test_changelog_output_not_installed():
    assert run('view-changelog curse:masque').stdout == ''


def test_changelog_output_argless():
    install_masque()
    assert run('view-changelog').stdout.startswith('curse:masque:')


def test_changelog_output_remote():
    assert run('view-changelog --remote curse:masque').stdout.startswith('curse:masque:')


def test_changelog_warning_if_no_pandoc():
    stderr = (
        ''
        if shutil.which('pandoc')
        else '! pandoc is not installed; changelogs will not be converted\n'
    )
    assert run('view-changelog curse:masque').stderr == stderr


@pytest.mark.parametrize(
    ('command', 'exit_code'),
    [
        ('list mas', 0),
        ('info foo', 0),
        ('reveal mas', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(
    monkeypatch: pytest.MonkeyPatch,
    command: str,
    exit_code: int,
):
    install_masque()
    monkeypatch.setattr('instawow._utils.file.reveal_folder', mock.MagicMock())
    assert run(command).exit_code == exit_code


def test_can_list_with_substr_match():
    install_masque()
    assert run('list mas').stdout == 'curse:masque\n'
    assert run('list foo').stdout == ''
    assert run('list -f detailed mas').stdout.startswith('curse:masque')
    (masque,) = json.loads(run('list -f json mas').stdout)
    assert (masque['source'], masque['slug']) == ('curse', 'masque')


def test_json_export():
    install_masque()
    output = run('list -f json').stdout
    assert json.loads(output)[0]['name'] == 'Masque'


def test_show_version():
    from instawow._version import get_version

    assert run('--version').stdout == f'instawow, version {get_version()}\n'


def test_plugin_hook_command_can_be_invoked():
    pytest.importorskip('instawow_test_plugin')
    assert run('plugins foo').stdout == 'success!\n'


# Skip loading the `_iw_web_client_ctx` fixture
# to allow deleting the cache on Windows.
@pytest.mark.parametrize('_iw_web_client_ctx', [None])
def test_clear_cache(
    iw_profile_config: ProfileConfig,
):
    assert iw_profile_config.global_config.dirs.cache.is_dir()
    assert run('cache clear').exit_code == 0
    assert not iw_profile_config.global_config.dirs.cache.is_dir()



from __future__ import annotations

import json
import os
from collections.abc import Mapping
from functools import lru_cache, partial
from pathlib import Path
from typing import Any, Literal, NewType, TypedDict

import attrs
import cattrs
import cattrs.gen
import cattrs.preconf.json


class FieldMetadata(TypedDict, total=False):
    env_prefix: str | Literal[True]
    store: Literal[True, 'independently']
    store_alias: str


SecretStr = NewType('SecretStr', str)


class UninitialisedConfigError(Exception):
    pass


def ensure_dirs(*dirs: Path) -> None:
    for dir_ in dirs:
        dir_.mkdir(exist_ok=True, parents=True)


def make_config_converter() -> cattrs.Converter:
    converter = cattrs.Converter()
    cattrs.preconf.json.configure_converter(converter)
    converter.register_structure_hook(Path, lambda v, _: Path(v))
    converter.register_unstructure_hook(Path, str)
    return converter


@lru_cache(1)
def make_display_converter():
    import cattrs.gen

    converter = make_config_converter()

    @partial(converter.register_unstructure_hook, SecretStr)
    def _(value: SecretStr):
        return '*' * 10

    @converter.register_unstructure_hook_factory(attrs.has)
    def _[T](cls: type[T], converter: cattrs.Converter):
        return cattrs.gen.make_dict_unstructure_fn(cls, converter, _cattrs_include_init_false=True)

    return converter


@lru_cache(1)
def _make_write_converter():
    converter = make_config_converter()

    @converter.register_unstructure_hook_factory(attrs.has)
    def _[T](type_: type[T]):
        return cattrs.gen.make_dict_unstructure_fn(
            type_,
            converter,
            **{  # pyright: ignore[reportArgumentType]  # See microsoft/pyright#5255
                f.name: cattrs.gen.override(omit=True)
                for f in attrs.fields(type_)
                if not f.metadata.get('store')
            },
        )

    return converter


def write_config(config: object, config_path: Path) -> None:
    converter = _make_write_converter()
    config_path.write_text(
        json.dumps(converter.unstructure(config), indent=2),
        encoding='utf-8',
    )


def read_config(attrs_cls: type, config_path: Path, missing_ok: bool = False) -> dict[str, Any]:
    try:
        values = json.loads(config_path.read_bytes())
        for field in attrs.fields(
            attrs.resolve_types(attrs_cls),
        ):
            if not field.init:
                continue

            metadata: FieldMetadata = field.metadata
            store = metadata.get('store')
            name = metadata.get('store_alias') or field.name

            if store == 'independently':
                maybe_values = read_config(
                    field.type, config_path.with_stem(f'{config_path.stem}.{name}'), True
                )
                if maybe_values:
                    values |= {field.name: maybe_values}
            elif name != field.name and name in values:
                values[field.name] = values[name]

    except FileNotFoundError:
        if missing_ok:
            return {}
        raise UninitialisedConfigError from None
    else:
        return values


def _parse_env_var(field_type: type, value: str):
    if attrs.has(field_type):
        return json.loads(value)
    elif field_type is bool:
        return attrs.converters.to_bool(value)
    return value


def read_env_vars(
    attrs_cls: type, values: Mapping[str, Any], parent_env_prefix: str | None = None
) -> dict[str, Any]:
    def iter_read():
        for field in attrs.fields(
            attrs.resolve_types(attrs_cls),
        ):
            if not field.init:
                continue

            value = values.get(field.name, attrs.NOTHING)

            metadata: FieldMetadata = field.metadata
            env_prefix = metadata.get('env_prefix')

            if env_prefix:
                env_key = f'{parent_env_prefix if env_prefix is True else env_prefix}_{field.name}'.upper()
                env_value = os.environ.get(env_key, attrs.NOTHING)
                if env_value is not attrs.NOTHING:
                    value = _parse_env_var(field.type, env_value)
                elif attrs.has(field.type):
                    value = read_env_vars(
                        field.type,
                        value if value is not attrs.NOTHING else {},
                        f'{env_prefix}_{field.name}',
                    )

            if value is not attrs.NOTHING:
                yield field.name, value

    return dict(iter_read())


config_converter = make_config_converter()


@config_converter.register_structure_hook_factory(attrs.has)
def _[T](type_: type[T]):
    "Allow passing in a structured attrs instance to ``structure``."
    structure = config_converter.gen_structure_attrs_fromdict(type_)

    def structure_wrapper(value: Mapping[str, object], type_: type):
        return value if isinstance(value, type_) else structure(value, type_)

    return structure_wrapper



from __future__ import annotations

import os
from collections.abc import Sequence
from datetime import datetime, timedelta
from enum import IntEnum
from functools import partial
from typing import NotRequired

from typing_extensions import TypedDict
from yarl import URL

from .. import config_ctx, http, http_ctx
from .._logging import logger
from .._utils.aio import gather
from .._utils.iteration import uniq
from ..definitions import ChangelogFormat, Defn, SourceMetadata, Strategy
from ..progress_reporting import make_download_progress
from ..resolvers import (
    AccessToken,
    BaseResolver,
    CatalogueEntryCandidate,
    HeadersIntent,
    PkgCandidate,
)
from ..results import PkgFilesMissing, PkgFilesNotMatching, PkgNonexistent, resultify
from ..wow_installations import Flavour, get_compatible_flavours, to_flavourful_enum

_CF_WOW_GAME_ID = 1


class _CfCoreModLinks(TypedDict):
    websiteUrl: str
    wikiUrl: str
    issuesUrl: str
    sourceUrl: str


class _CfCoreModStatus(IntEnum):
    New = 1
    ChangesRequired = 2
    UnderSoftReview = 3
    Approved = 4
    Rejected = 5
    ChangesMade = 6
    Inactive = 7
    Abandoned = 8
    Deleted = 9
    UnderReview = 10


class _CfCoreCategory(TypedDict):
    id: int
    gameId: int
    name: str
    slug: str
    url: str
    iconUrl: str
    dateModified: str  # date-time
    isClass: bool
    classId: int
    parentCategoryId: int


class _CfCoreModAuthor(TypedDict):
    id: int
    name: str
    url: str


class _CfCoreModAsset(TypedDict):
    id: int
    modId: int
    title: str
    description: str
    thumbnailUrl: str
    url: str


class _CfCoreFileIndex(TypedDict):
    gameVersion: str
    fileId: int
    filename: str
    releaseType: _CfCoreFileReleaseType
    gameVersionTypeId: int
    modLoader: int


class _CfCoreFileStatus(IntEnum):
    Processing = 1
    ChangesRequired = 2
    UnderReview = 3
    Approved = 4
    Rejected = 5
    MalwareDetected = 6
    Deleted = 7
    Archived = 8
    Testing = 9
    Released = 10
    ReadyForReview = 11
    Deprecated = 12
    Baking = 13
    AwaitingPublishing = 14
    FailedPublishing = 15


class _CfCoreFileReleaseType(IntEnum):
    Release = 1
    Beta = 2
    Alpha = 3


class _CfCoreHashAlgo(IntEnum):
    Sha1 = 1
    Md5 = 2


class _CfCoreFileHash(TypedDict):
    value: str
    algo: _CfCoreHashAlgo


class _CfCoreSortableGameVersionTypeId(IntEnum):
    "Extracted from https://api.curseforge.com/v1/games/1/version-types."

    Retail = 517
    VanillaClassic = 67408
    TbcClassic = 73246
    WrathClassic = 73713
    CataClassic = 77522
    MistsClassic = 79434


class _CfCoreSortableGameVersion(TypedDict):
    gameVersionName: str
    gameVersionPadded: str
    gameVersion: str
    gameVersionReleaseDate: str  # date-time
    gameVersionTypeId: _CfCoreSortableGameVersionTypeId


class _CfCoreFileRelationType(IntEnum):
    EmbeddedLibrary = 1
    OptionalDependency = 2
    RequiredDependency = 3
    Tool = 4
    Incompatible = 5
    Include = 6


class _CfCoreFileDependency(TypedDict):
    modId: int
    fileId: int
    relationType: _CfCoreFileRelationType


class _CfCoreFileModule(TypedDict):
    name: str
    fingerprint: int


class _CfCoreFile(TypedDict):
    id: int
    gameId: int
    modId: int
    isAvailable: bool
    displayName: str
    fileName: str
    releaseType: _CfCoreFileReleaseType
    fileStatus: _CfCoreFileStatus
    hashes: list[_CfCoreFileHash]
    fileDate: str  # date-time
    fileLength: int
    downloadCount: int
    downloadUrl: str | None  # null if distribution is forbidden
    gameVersions: list[str]
    sortableGameVersions: list[_CfCoreSortableGameVersion]
    dependencies: list[_CfCoreFileDependency]
    exposeAsAlternative: NotRequired[bool]
    parentProjectFileId: NotRequired[int]
    alternateFileId: int
    isServerPack: bool
    serverPackFileId: NotRequired[int]
    fileFingerprint: int
    modules: list[_CfCoreFileModule]


class _CfCoreMod(TypedDict):
    id: int
    gameId: int
    name: str
    slug: str
    links: _CfCoreModLinks
    summary: str
    status: _CfCoreModStatus
    downloadCount: int
    isFeatured: bool
    primaryCategoryId: int
    categories: list[_CfCoreCategory]
    authors: list[_CfCoreModAuthor]
    logo: _CfCoreModAsset
    screenshots: list[_CfCoreModAsset]
    mainFileId: int
    latestFiles: list[_CfCoreFile]
    latestFilesIndexes: list[_CfCoreFileIndex]
    dateCreated: str  # date-time
    dateModified: str  # date-time
    dateReleased: str  # date-time
    allowModDistribution: bool | None


class _CfCoreModsSearchSortField(IntEnum):
    Featured = 1
    Popularity = 2
    LastUpdated = 3
    Name = 4
    Author = 5
    TotalDownloads = 6
    Category = 7
    GameVersion = 8
    EarlyAccess = 9
    FeaturedReleased = 10
    ReleasedDate = 11
    Rating = 12


class _CfCoreResponsePagination(TypedDict):
    index: int
    pageSize: int
    resultCount: int
    totalCount: int | None


class _CfCoreDataResponse[T](TypedDict):
    data: T


class _CfCorePaginatedDataResponse[T](TypedDict):
    data: T
    pagination: _CfCoreResponsePagination


_ALTERNATIVE_API_URL_ENV_KEY = 'INSTAWOW_CF_API_URL'
_alternative_api_url = os.environ.get(_ALTERNATIVE_API_URL_ENV_KEY)


_VERSION_SEP = '_'


class CfCoreResolver(BaseResolver[_CfCoreMod]):
    metadata = SourceMetadata(
        id='curse',
        name='CurseForge',
        strategies=frozenset(
            {
                Strategy.AnyFlavour,
                Strategy.AnyReleaseType,
                Strategy.VersionEq,
            }
        ),
        changelog_format=ChangelogFormat.Html,
        addon_toc_key='X-Curse-Project-ID',
    )

    # Ref: https://docs.curseforge.com/
    __mod_api_url = URL(_alternative_api_url or 'https://api.curseforge.com/v1').joinpath('mods')

    @AccessToken
    def access_token():
        return config_ctx.config().global_config.access_tokens.cfcore, not _alternative_api_url

    def get_alias_from_url(self, url: URL):
        if (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    def make_request_headers(self, intent: HeadersIntent | None = None):
        if not _alternative_api_url and intent is not HeadersIntent.Download:
            maybe_access_token = self.access_token.get()
            if maybe_access_token:
                return {'x-api-key': maybe_access_token}

    async def resolve(self, defns: Sequence[Defn]):
        defn_ids = {d: d.id or d.alias for d in defns}
        numeric_ids = uniq(i for i in defn_ids.values() if i.isdigit())
        if not numeric_ids:
            return await super().resolve(defns)  # Fast path.

        async with http_ctx.web_client().post(
            self.__mod_api_url,
            expire_after=timedelta(minutes=5),
            headers=self.make_request_headers(),
            json={'modIds': numeric_ids},
            raise_for_status=True,
        ) as response:
            response_json: _CfCoreDataResponse[list[_CfCoreMod]] = await response.json()

        addons_by_id = {str(r['id']): r for r in response_json['data']}

        resolve_one = resultify(self.resolve_one)
        results = await gather(resolve_one(d, addons_by_id.get(i)) for d, i in defn_ids.items())
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _CfCoreMod | None):
        if metadata is None:
            if defn.alias.isdigit():
                async with http_ctx.web_client().get(
                    self.__mod_api_url / defn.alias,
                    expire_after=timedelta(minutes=15),
                    headers=self.make_request_headers(),
                ) as mod_response:
                    if mod_response.status == 404:
                        raise PkgNonexistent

                    mod_response.raise_for_status()

                    metadata = await mod_response.json()
                    assert metadata

            else:
                async with http_ctx.web_client().get(
                    (self.__mod_api_url / 'search').with_query(
                        gameId=_CF_WOW_GAME_ID, slug=defn.alias
                    ),
                    expire_after=timedelta(minutes=15),
                    headers=self.make_request_headers(),
                    raise_for_status=True,
                ) as mod_response:
                    mod_response_json: _CfCorePaginatedDataResponse[
                        list[_CfCoreMod]
                    ] = await mod_response.json()
                    if not mod_response_json['data']:
                        raise PkgNonexistent

                    [metadata] = mod_response_json['data']

        version_eq = defn.strategies[Strategy.VersionEq]
        if version_eq:
            _, file_sep, file_id = version_eq.rpartition(_VERSION_SEP)
            files_url = self.__mod_api_url / str(metadata['id']) / 'files'
            if file_sep and file_id:
                files_url /= file_id

            async with http_ctx.web_client().get(
                files_url,
                expire_after=timedelta(hours=1),
                headers=self.make_request_headers(),
                raise_for_status=True,
                trace_request_ctx={
                    'progress': make_download_progress(
                        label=f'Fetching metadata from {self.metadata.name}'
                    )
                },
            ) as files_response:
                files_response_json: (
                    _CfCoreDataResponse[_CfCoreFile] | _CfCoreDataResponse[list[_CfCoreFile]]
                ) = await files_response.json()

            files = files_response_json['data']
            if isinstance(files, list):
                files = next(
                    ([f] for f in files if f['displayName'] == version_eq), list[_CfCoreFile]()
                )
            else:
                files = [files]

        else:
            files = metadata['latestFiles']

        if not files:
            raise PkgFilesMissing

        files = [f for f in files if not f.get('exposeAsAlternative')]

        # Allow pre-releases only if no stable releases exist or explicitly opted into.
        if not defn.strategies[Strategy.AnyReleaseType] and any(
            f['releaseType'] == _CfCoreFileReleaseType.Release for f in files
        ):
            files = [f for f in files if f['releaseType'] == _CfCoreFileReleaseType.Release]

        desired_flavours = get_compatible_flavours(
            config_ctx.config().track, defn.strategies[Strategy.AnyFlavour]
        )
        for desired_flavour in desired_flavours:
            if desired_flavour:
                type_id = to_flavourful_enum(desired_flavour, _CfCoreSortableGameVersionTypeId)
                shortlisted_files = (
                    f
                    for f in files
                    if any(s['gameVersionTypeId'] == type_id for s in f['sortableGameVersions'])
                )
            else:
                shortlisted_files = files

            try:
                file = max(
                    shortlisted_files,
                    # The ``id`` is just a counter so we don't have to go digging around dates
                    key=lambda f: f['id'],
                )
            except ValueError:
                continue
            break

        else:
            raise PkgFilesNotMatching(defn.strategies)

        if file['downloadUrl'] is None:
            raise (
                PkgFilesMissing('package distribution is forbidden')
                if metadata['allowModDistribution'] is False
                else PkgFilesMissing
            )

        return PkgCandidate(
            id=str(metadata['id']),
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['links']['websiteUrl'],
            download_url=file['downloadUrl'],
            date_published=datetime.fromisoformat(file['fileDate']),
            version=f'{file["displayName"]}{_VERSION_SEP}{file["id"]}',
            changelog_url=str(
                self.__mod_api_url / f'{metadata["id"]}/files/{file["id"]}/changelog'
            ),
            deps=[
                {'id': str(d['modId'])}
                for d in file['dependencies']
                if d['relationType'] == _CfCoreFileRelationType.RequiredDependency
            ],
        )

    async def get_changelog(self, uri: URL):
        async with http_ctx.web_client().get(
            uri,
            expire_after=http.CACHE_INDEFINITELY,
            headers=self.make_request_headers(),
            raise_for_status=True,
        ) as response:
            response_json: _CfCoreDataResponse[str] = await response.json()
            return response_json['data']

    async def catalogue(self):
        from aiohttp import ClientTimeout

        supported_flavours = [
            (f, to_flavourful_enum(f, _CfCoreSortableGameVersionTypeId)) for f in Flavour
        ]

        def excise_flavours(files: list[_CfCoreFile]):
            for flavour, version_type in supported_flavours:
                if any(
                    not f.get('exposeAsAlternative', False)
                    and any(
                        s['gameVersionTypeId'] == version_type for s in f['sortableGameVersions']
                    )
                    for f in files
                ):
                    yield flavour

        def excise_folders(files: list[_CfCoreFile]):
            return uniq(frozenset(m['name'] for m in f['modules']) for f in files)

        STEP = 50
        MAX_OFFSET = 10_000  # The CF API craps out after 9,999 results

        get = partial(
            http_ctx.web_client().get,
            raise_for_status=True,
            timeout=ClientTimeout(total=10),
        )
        access_token = self.access_token.get()
        if access_token:
            get = partial(get, headers={'x-api-key': access_token})

        for offset in range(0, MAX_OFFSET, STEP):
            url = (self.__mod_api_url / 'search').with_query(
                gameId=_CF_WOW_GAME_ID,
                sortField=_CfCoreModsSearchSortField.LastUpdated,
                sortOrder='desc',
                pageSize=STEP,
                index=offset,
            )
            logger.debug(f'Retrieving {url}')

            for attempt in range(3):
                try:
                    async with get(url) as response:
                        response_json: _CfCorePaginatedDataResponse[
                            list[_CfCoreMod]
                        ] = await response.json()
                        break
                except TimeoutError:
                    logger.debug(f'Request timed out; attempt {attempt + 1} of 3')
            else:
                raise RuntimeError('Maximum number of attempts exceeded')

            items = response_json['data']
            if not items:
                break

            for item in items:
                yield CatalogueEntryCandidate(
                    id=str(item['id']),
                    slug=item['slug'],
                    name=item['name'],
                    url=item['links']['websiteUrl'],
                    game_flavours=frozenset(excise_flavours(item['latestFiles'])),
                    download_count=item['downloadCount'],
                    last_updated=datetime.fromisoformat(item['dateReleased']),
                    folders=excise_folders(item['latestFiles']),
                )



from __future__ import annotations

import importlib.util
import sys


def make_getattr(package_name: str):
    def __getattr__(name: str):
        """Defer importing own modules until attempting to access an attribute.

        Importing this in ``__init__`` will overwrite relative imports.
        """
        fullname = package_name + '.' + name

        try:
            return sys.modules[fullname]
        except KeyError:
            spec = importlib.util.find_spec(fullname)
            if spec is None or spec.loader is None:
                # ``AttributeError`` is converted to an ``ImportError`` by the import machinery.
                raise AttributeError from None

            spec.loader = loader = importlib.util.LazyLoader(spec.loader)
            sys.modules[fullname] = module = importlib.util.module_from_spec(spec)
            loader.exec_module(module)

            return module

    return __getattr__



from __future__ import annotations

import pytest
from yarl import URL

from instawow._sources.wowi import WowiResolver
from instawow.definitions import Defn

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


@pytest.fixture
def wowi_resolver():
    return WowiResolver()


async def test_resolve_addon(
    wowi_resolver: WowiResolver,
):
    defn = Defn('wowi', '12097-masque')

    result = (await wowi_resolver.resolve([defn]))[defn]
    assert type(result) is dict


async def test_changelog_url_format(
    wowi_resolver: WowiResolver,
):
    defn = Defn('wowi', '12097-masque')

    result = (await wowi_resolver.resolve([defn]))[defn]
    assert type(result) is dict
    assert result['changelog_url'].startswith('data:,')


@pytest.mark.parametrize(
    ('url', 'extracted_alias'),
    [
        (
            'https://www.wowinterface.com/downloads/landing.php?fileid=12097',
            '12097',
        ),
        (
            'https://wowinterface.com/downloads/landing.php?fileid=12097',
            '12097',
        ),
        (
            'https://www.wowinterface.com/downloads/fileinfo.php?id=12097',
            '12097',
        ),
        (
            'https://wowinterface.com/downloads/fileinfo.php?id=12097',
            '12097',
        ),
        (
            'https://www.wowinterface.com/downloads/download12097-Masque',
            '12097',
        ),
        (
            'https://wowinterface.com/downloads/download12097-Masque',
            '12097',
        ),
        (
            'https://www.wowinterface.com/downloads/info12097-Masque.html',
            '12097',
        ),
        (
            'https://wowinterface.com/downloads/info12097-Masque.html',
            '12097',
        ),
        (
            'https://www.wowinterface.com/downloads/info12097',
            '12097',
        ),
        (
            'https://wowinterface.com/downloads/info12097',
            '12097',
        ),
    ],
)
def test_can_extract_alias_from_url(
    wowi_resolver: WowiResolver,
    url: str,
    extracted_alias: str,
):
    assert wowi_resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import lru_cache, partial
from typing import TYPE_CHECKING

import attrs

type _Validator[T] = Callable[[object, attrs.Attribute[T], T], None]


fauxfrozen = attrs.frozen if TYPE_CHECKING else partial(attrs.define, unsafe_hash=True)


def enrich_validator_exc[T](validator: _Validator[T]) -> _Validator[T]:
    "Pretend validation error originates from cattrs for uniformity with structural errors."

    import cattrs

    def wrapper(model: object, attr: attrs.Attribute[T], value: T):
        try:
            validator(model, attr, value)
        except BaseException as exc:
            note = f'Structuring class {model.__class__.__name__} @ attribute {attr.name}'
            exc.add_note(cattrs.AttributeValidationNote(note, attr.name, attr.type))
            raise

    return wrapper


@fauxfrozen
class EvolveIdent:
    value: object


def evolve[T](attrs_instance: T, changes: Mapping[str, object | EvolveIdent]) -> T:
    return attrs.evolve(
        attrs_instance,
        **{
            k: evolve(
                getattr(attrs_instance, k),
                v,  # pyright: ignore[reportUnknownArgumentType]
            )
            if isinstance(v, Mapping)
            else v.value
            if isinstance(v, EvolveIdent)
            else v
            for k, v in changes.items()
        },
    )


@lru_cache(1)
def simple_converter():
    from cattrs import Converter

    return Converter()
